import logging

# import yasfpy.log as log
from time import time

import numpy as np
from math import ceil
from numba import cuda
from scipy.sparse.linalg import LinearOperator

# from scipy.spatial.distance import pdist, squareform
from scipy.special import spherical_jn, spherical_yn

# from scipy.special import hankel1
# from scipy.special import lpmv

from yasfpy.parameters import Parameters
from yasfpy.numerics import Numerics
from yasfpy.functions.spherical_functions_trigon import spherical_functions_trigon
from yasfpy.functions.t_entry import t_entry

from yasfpy.functions.misc import transformation_coefficients
from yasfpy.functions.misc import multi2single_index
from yasfpy.functions.misc import mutual_lookup

from yasfpy.functions.cpu_numba import compute_idx_lookups
from yasfpy.functions.cpu_numba import particle_interaction, compute_field
from yasfpy.functions.cuda_numba import particle_interaction_gpu, compute_field_gpu


class Simulation:
    """This class represents the simulation of YASF (Yet Another Scattering Framework).
    It contains methods for initializing the simulation, computing lookup tables, and calculating mie coefficients.
    """

    def __init__(self, parameters: Parameters, numerics: Numerics):
        """
        Initialize the Simulation object.

        Args:
            parameters (Parameters): The parameters for the simulation.
            numerics (Numerics): The numerics for the simulation.
        """
        self.parameters = parameters
        self.numerics = numerics

        # self.log = log.infoing_logger(__name__)
        self.log = logging.getLogger(self.__class__.__module__)
        self.__setup()

    def legacy_compute_lookup_particle_distances(self):
        """
        The largest distance between two particles is divided into segments provided by `Numerics.particle_distance_resolution`.
        This array is then used as a lookup for the calculation of the spherical Hankel function.

        Notes
        -----
        This function has been ported from the Matlab Celes framework but is not used by YASF!
        """
        # add two zeros at the beginning to allow interpolation
        # also in the first segment
        step = self.numerics.particle_distance_resolution
        maxdist = (
            self.parameters.particles.max_particle_distance
            + 3 * self.numerics.particle_distance_resolution
        )
        self.lookup_particle_distances = np.concatenate(
            (np.array([0]), np.arange(0, maxdist + np.finfo(float).eps, step))
        )

    def legacy_compute_h3_table(self):
        """
        Computes the spherical hankel function
        at the points calculated in `Simulation.legacy_compute_lookup_particle_distances()`.

        Attributes:
            h3_table (np.ndarray): Lookup table of the spherical hankel function values at `self.lookup_particle_distances`

        Notes:
            This function has been ported from the Matlab Celes framework but is not used by YASF!
        """
        self.h3_table = np.zeros(
            (
                2 * self.numerics.lmax + 1,
                self.lookup_particle_distances.shape[0],
                self.parameters.medium_refractive_index.shape[0],
            ),
            dtype=complex,
        )
        size_param = np.outer(self.lookup_particle_distances, self.parameters.k_medium)

        for p in range(2 * self.numerics.lmax + 1):
            self.h3_table[p, :, :] = spherical_jn(p, size_param) + 1j * spherical_yn(
                p, size_param
            )

    def __compute_idx_lookup(self):
        """
        Creates a lookup table with the indices used in further calculations.
        The lookup table is created using `compute_idx_lookups` function from `yasfpy.functions.cpu_numba`.

        Attributes:
            idx_lookup (np.ndarray): Lookup table of the indices to iterate over large arrays.

        Notes:
            This function utilizes Numba to optimize the computations.
        """
        self.idx_lookup = compute_idx_lookups(
            self.numerics.lmax, self.parameters.particles.number
        )

    def __compute_lookups(self):
        """
        Computes various lookup tables for each particle.

        Attributes:
            sph_j (np.ndarray): Spherical Bessel function lookup table calculated for pair-wise particle distances.
            sph_h (np.ndarray): Spherical Hankel function lookup table calculated for pair-wise particle distances.
            plm (np.ndarray): Associated Legendre polynomial lookup table calculated for the cosine value of the pairwise particle inclination angles.
            e_j_dm_phi (np.ndarray): Exponential function lookup table calculated for the pairwise particle azimuthal angles.

        Notes:
            This function uses numba (https://numba.pydata.org/) under the hood to speed up the computations.
        """
        lookup_computation_time_start = time()
        # TODO: new, could be error prone and is not tested yet!
        self.sph_j, self.sph_h, self.e_j_dm_phi, self.plm = mutual_lookup(
            self.numerics.lmax,
            self.parameters.particles.position,
            self.parameters.particles.position,
            self.parameters.k_medium,
        )[:4]

        # lmax = self.numerics.lmax
        # particle_number = self.parameters.particles.number

        # dists = squareform(pdist(self.parameters.particles.position))
        # ct = np.divide(
        #   np.subtract.outer(
        #     self.parameters.particles.position[:, 2], self.parameters.particles.position[:, 2]),
        #   dists,
        #   out = np.zeros((particle_number, particle_number)),
        #   where = dists != 0)
        # phi = np.arctan2(
        #   np.subtract.outer(
        #     self.parameters.particles.position[:, 1], self.parameters.particles.position[:, 1]),
        #   np.subtract.outer(self.parameters.particles.position[:, 0], self.parameters.particles.position[:, 0]))

        # size_param = np.outer(dists.ravel(), self.parameters.k_medium).reshape(
        #   [particle_number, particle_number, self.parameters.k_medium.shape[0]])

        # self.sph_h = np.zeros((2 * lmax + 1, particle_number, particle_number, self.parameters.k_medium.shape[0]), dtype=complex)
        # self.sph_j = np.zeros_like(self.sph_h)
        # self.e_j_dm_phi = np.zeros((4 * lmax + 1, particle_number, particle_number), dtype=complex)
        # self.plm = np.zeros(((lmax + 1) * (2 * lmax + 1),
        #           particle_number, particle_number))

        # for p in range(2 * lmax + 1):
        #   self.sph_h[p, :, :, :] = np.sqrt(
        #     np.divide(
        #       np.pi / 2,
        #       size_param,
        #       out=np.zeros_like(size_param),
        #       where=size_param != 0)
        #   ) * hankel1(p + 1/2, size_param)
        #   self.sph_j[p, :, :, :] = spherical_jn(p, size_param)
        #   self.e_j_dm_phi[p, :, :] = np.exp(1j * (p - 2 * lmax) * phi)
        #   self.e_j_dm_phi[p + 2 * lmax, :, :] = np.exp(1j * p * phi)
        #   for absdm in range(p + 1):
        #     cml = np.sqrt((2 * p + 1) / 2 /
        #             np.prod(np.arange(p - absdm + 1, p + absdm + 1)))
        #     self.plm[p * (p + 1) // 2 + absdm, :, :] = cml * \
        #       np.power(-1.0, absdm) * lpmv(absdm, p, ct)

        # self.sph_h = np.nan_to_num(
        #   self.sph_h, nan=0) + np.isnan(self.sph_h) * 1

        lookup_computation_time_stop = time()
        self.log.info(
            "Computing lookup tables took %f s"
            % (lookup_computation_time_stop - lookup_computation_time_start)
        )

    def __setup(self):
        """
        An internal setup function called upon object creation.
        The following functions are called:

        - [__compute_idx_lookups][simulation.Simulation.__compute_idx_lookup]
        - [__compute_lookups][simulation.Simulation.__compute_lookups]
        """
        self.__compute_idx_lookup()
        self.__compute_lookups()

    def compute_mie_coefficients(self):
        """
        Computes the mie coefficients for the unique pair
        of particle radius and the refractive index of the particle.

        Attributes:
            mie_coefficients (np.ndarray): Mie coefficients table

        See Also:
            [t_entry][functions.t_entry.t_entry] : T-Matrix entry function

        Notes:
            Due to the four nested loops (particles, tau, l, and m),
            it could be rewritten using `numba` to speed the process up.
        """
        self.mie_coefficients = np.zeros(
            (
                self.parameters.particles.num_unique_pairs,
                self.numerics.nmax,
                self.parameters.wavelength.shape[0],
            ),
            dtype=complex,
        )

        self.scatter_to_internal = np.zeros_like(self.mie_coefficients)

        for u_i in range(self.parameters.particles.num_unique_pairs):
            for tau in range(1, 3):
                for l in range(1, self.numerics.lmax + 1):
                    for m in range(-l, l + 1):
                        jmult = multi2single_index(0, tau, l, m, self.numerics.lmax)
                        self.mie_coefficients[u_i, jmult, :] = t_entry(
                            tau=tau,
                            l=l,
                            k_medium=self.parameters.k_medium,
                            k_sphere=self.parameters.omega
                            * self.parameters.particles.unique_radius_index_pairs[
                                u_i, 1:
                            ],
                            radius=np.real(
                                self.parameters.particles.unique_radius_index_pairs[
                                    u_i, 0
                                ]
                            ),
                        )

                        self.scatter_to_internal[u_i, jmult, :] = t_entry(
                            tau=tau,
                            l=l,
                            k_medium=self.parameters.k_medium,
                            k_sphere=self.parameters.omega
                            * self.parameters.particles.unique_radius_index_pairs[
                                u_i, 1:
                            ],
                            radius=np.real(
                                self.parameters.particles.unique_radius_index_pairs[
                                    u_i, 0
                                ]
                            ),
                            field_type="ratio",
                        )

    def compute_initial_field_coefficients(self):
        r"""
        Computes initial field coefficients $a_{\\tau ,l,m}$ and $b_{\\tau ,l,m}$.
        Depending on the `beam_width`, one of two functions is called:

        - [__compute_initial_field_coefficients_wavebundle_normal_incidence][simulation.Simulation.__compute_initial_field_coefficients_wavebundle_normal_incidence], $\\text{beam width} \\in (0, \\infty)$
        - [__compute_initial_field_coefficients_planewave][simulation.Simulation.__compute_initial_field_coefficients_planewave], $\\text{beam width} = 0$ or $\\text{beam width} = \\infty$

        Attributes:
            initial_field_coefficients (np.ndarray): Initial field coefficients
        """
        self.log.info("compute initial field coefficients ...")

        if np.isfinite(self.parameters.initial_field.beam_width) and (
            self.parameters.initial_field.beam_width > 0
        ):
            self.log.info("\t Gaussian beam ...")
            if self.parameters.initial_field.normal_incidence:
                self.__compute_initial_field_coefficients_wavebundle_normal_incidence()
            else:
                self.log.error("\t this case is not implemented")
        else:
            self.log.info("\t plane wave ...")
            self.__compute_initial_field_coefficients_planewave()

        self.log.info("done")

    def compute_right_hand_side(self):
        r"""
        Computes the right hand side $T \\cdot a_I$ of the equation $M \\cdot b = T \\cdot a_I$.

        Attributes
        ----------
        right_hand_side : np.ndarray
            Right hand side of the equation $M \\cdot b = T \\cdot a_I$

        Notes
        -----
        For more information regarding the equation, please refer to the paper by Celes (https://arxiv.org/abs/1706.02145).
        """
        self.right_hand_side = (
            self.mie_coefficients[self.parameters.particles.single_unique_array_idx, :]
            * self.initial_field_coefficients
        )

    def __compute_initial_field_coefficients_planewave(self):
        """The function computes the initial field coefficients for a plane wave based on given parameters
        and spherical coordinates.

        """
        lmax = self.numerics.lmax
        E0 = self.parameters.initial_field.amplitude
        k = self.parameters.k_medium

        beta = self.parameters.initial_field.polar_angle
        cb = np.cos(beta)
        sb = np.sin(beta)
        alpha = self.parameters.initial_field.azimuthal_angle

        # pi and tau symbols for transformation matrix B_dagger
        pilm, taulm = spherical_functions_trigon(lmax, beta)

        # cylindrical coordinates for relative particle positions
        relative_particle_positions = (
            self.parameters.particles.position
            - self.parameters.initial_field.focal_point
        )
        kvec = np.outer(np.array((sb * np.cos(alpha), sb * np.sin(alpha), cb)), k)
        eikr = np.exp(1j * np.matmul(relative_particle_positions, kvec))

        # clean up some memory?
        del (k, beta, cb, sb, kvec, relative_particle_positions)

        self.initial_field_coefficients = np.zeros(
            (
                self.parameters.particles.number,
                self.numerics.nmax,
                self.parameters.k_medium.size,
            ),
            dtype=complex,
        )
        for m in range(-lmax, lmax + 1):
            for tau in range(1, 3):
                for l in range(np.max([1, np.abs(m)]), lmax + 1):
                    n = multi2single_index(0, tau, l, m, lmax)
                    self.initial_field_coefficients[:, n, :] = (
                        4
                        * E0
                        * np.exp(-1j * m * alpha)
                        * eikr
                        * transformation_coefficients(
                            pilm,
                            taulm,
                            tau,
                            l,
                            m,
                            self.parameters.initial_field.pol,
                            dagger=True,
                        )
                    )

    def __compute_initial_field_coefficients_wavebundle_normal_incidence(self):
        """The function initializes the field coefficients for a wave bundle incident at normal incidence.

        TODO:
            Implement this function using the celes function [initial_field_coefficients_wavebundle_normal_incidence.m](https://github.com/disordered-photonics/celes/blob/master/src/initial/initial_field_coefficients_wavebundle_normal_incidence.m)
        """
        self.initial_field_coefficients = (
            np.zeros(
                (
                    self.parameters.particles.number,
                    self.numerics.nmax,
                    self.parameters.k_medium.size,
                ),
                dtype=complex,
            )
            * np.nan
        )

    def coupling_matrix_multiply(self, x: np.ndarray, idx: int = None):
        """Computes the coupling matrix `wx` based on the input parameters.

        Args:
            x (np.ndarray): An input array of shape (n,) or (n, m), where n is the number of particles and m is the number
                of features for each particle. This array represents the input data for which the coupling
                matrix needs to be computed.
            idx (int): An optional integer that specifies the index of a specific spherical harmonic mode. If `idx` is provided,
                the computation will only be performed for that specific mode. If `idx` is not provided or set to `None`,
                the computation will be performed for all spherical harmonic modes.

        Returns:
            wx (np.ndarray): An array of shape (n, m, p), where n is the number of particles, m is the number of features for each
                particle, and p is the number of wavelengths. It represents the coupling matrix `wx`.
        """
        self.log.debug("prepare particle coupling ... ")
        preparation_time = time()

        lmax = self.numerics.lmax
        particle_number = self.parameters.particles.number
        jmax = particle_number * 2 * lmax * (lmax + 2)
        wavelengths_size = self.parameters.k_medium.shape[0]
        translation_table = self.numerics.translation_ab5
        associated_legendre_lookup = self.plm
        spherical_hankel_lookup = self.sph_h
        e_j_dm_phi_loopup = self.e_j_dm_phi

        idx_lookup = self.idx_lookup

        if idx is not None:
            spherical_hankel_lookup = spherical_hankel_lookup[:, :, :, idx]
            spherical_hankel_lookup = np.copy(
                spherical_hankel_lookup[:, :, :, np.newaxis]
            )
            wavelengths_size = 1

        self.log.debug("\t Starting Wx computation")
        if self.numerics.gpu:
            wx_real = np.zeros(x.shape + (wavelengths_size,), dtype=float)
            wx_imag = np.zeros_like(wx_real)

            idx_device = cuda.to_device(idx_lookup)
            x_device = cuda.to_device(x)
            wx_real_device = cuda.to_device(wx_real)
            wx_imag_device = cuda.to_device(wx_imag)
            translation_device = cuda.to_device(translation_table)
            associated_legendre_device = cuda.to_device(associated_legendre_lookup)
            spherical_hankel_device = cuda.to_device(spherical_hankel_lookup)
            e_j_dm_phi_device = cuda.to_device(e_j_dm_phi_loopup)

            threads_per_block = (16, 16, 2)
            blocks_per_grid_x = ceil(jmax / threads_per_block[0])
            blocks_per_grid_y = ceil(jmax / threads_per_block[1])
            blocks_per_grid_z = ceil(wavelengths_size / threads_per_block[2])
            blocks_per_grid = (blocks_per_grid_x, blocks_per_grid_y, blocks_per_grid_z)

            coupling_matrix_time = time()
            particle_interaction_gpu[blocks_per_grid, threads_per_block](
                lmax,
                particle_number,
                idx_device,
                x_device,
                wx_real_device,
                wx_imag_device,
                translation_device,
                associated_legendre_device,
                spherical_hankel_device,
                e_j_dm_phi_device,
            )
            wx_real = wx_real_device.copy_to_host()
            wx_imag = wx_imag_device.copy_to_host()
            wx = wx_real + 1j * wx_imag
            # particle_interaction.parallel_diagnostics(level=4)
            time_end = time()
            self.log.debug(
                "\t Time taken for preparation: %f"
                % (coupling_matrix_time - preparation_time)
            )
            self.log.debug(
                "\t Time taken for coupling matrix: %f"
                % (time_end - coupling_matrix_time)
            )
        else:
            # from numba_progress import ProgressBar
            # num_iterations = jmax * jmax * wavelengths
            # progress = ProgressBar(total=num_iterations)
            # progress = None
            wx = particle_interaction(
                lmax,
                particle_number,
                idx_lookup,
                x,
                translation_table,
                associated_legendre_lookup,
                spherical_hankel_lookup,
                e_j_dm_phi_loopup,
            )
            time_end = time()
            self.log.debug(
                "\t Time taken for coupling matrix: %f" % (time_end - preparation_time)
            )

        if idx is not None:
            wx = np.squeeze(wx)

        return wx

    def master_matrix_multiply(self, value: np.ndarray, idx: int):
        """Applies a T-matrix to a given value and returns the result.

        Args:
            value (np.ndarray): The input value for the matrix multiplication operation.
            idx (int): The index of the matrix to be multiplied.

        Returns:
            mx (np.ndarray): The result of the matrix multiplication operation.

        """
        wx = self.coupling_matrix_multiply(value, idx)

        self.log.debug("apply T-matrix ...")
        t_matrix_start = time()

        twx = (
            self.mie_coefficients[
                self.parameters.particles.single_unique_array_idx, :, idx
            ].ravel(order="C")
            * wx
        )
        mx = value - twx

        t_matrix_stop = time()
        self.log.debug(f"\t done in {t_matrix_stop - t_matrix_start} seconds.")

        return mx

    def compute_scattered_field_coefficients(self, guess: np.ndarray = None):
        """The function computes the scattered field coefficients using a linear operator and a solver.

        Args:
            guess (np.ndarray): Optional. The initial guess for the solution of the linear system. If no guess is provided,
                the `right_hand_side` variable is used as the initial guess.

        """
        self.log.info("compute scattered field coefficients ...")
        jmax = self.parameters.particles.number * self.numerics.nmax
        self.scattered_field_coefficients = np.zeros_like(
            self.initial_field_coefficients
        )
        self.scattered_field_err_codes = np.zeros(self.parameters.wavelengths_number)
        if guess is None:
            guess = self.right_hand_side
        for w in range(self.parameters.wavelengths_number):
            # def mmm(x):
            #     return self.master_matrix_multiply(x, w)

            # A = LinearOperator(shape=(jmax, jmax), matvec=mmm)

            A = LinearOperator(
                shape=(jmax, jmax), matvec=lambda x: self.master_matrix_multiply(x, w)
            )
            b = self.right_hand_side[:, :, w].ravel()
            x0 = guess[:, :, w].ravel()
            self.log.info(
                "Solver run %d/%d" % (w + 1, self.parameters.wavelengths_number)
            )
            x, err_code = self.numerics.solver.run(A, b, x0)
            self.scattered_field_coefficients[:, :, w] = x.reshape(
                self.right_hand_side.shape[:2]
            )
            self.scattered_field_err_codes[w] = err_code

    def compute_fields(self, sampling_points: np.ndarray):
        """The function `compute_fields` calculates the field at given sampling points using either CPU or
        GPU computation.

        Args:
            sampling_points (np.ndarray): The numpy array that represents the coordinates of the sampling points.
                It should have a shape of `(n, 3)`, where `n` is the number of sampling points and each row
                represents the `(x, y, z)` coordinates of a point.
        """
        if sampling_points.shape[0] < 1:
            self.log.error("Number of sampling points must be bigger than zero!")
            return
        if sampling_points.shape[1] != 3:
            self.log.error("The points have to have three coordinates (x,y,z)!")
            return

        # scatter_to_internal_table = np.sum((self.parameters.particles.position[:, np.newaxis, :] - sampling_points[np.newaxis, :, :])**2, axis = 2)
        # scatter_to_internal_table = scatter_to_internal_table < self.parameters.particles.r[:, np.newaxis]**2

        self.log.info("Computing mutual lookup")
        lookup_computation_time_start = time()
        (
            _,
            sph_h,
            e_j_dm_phi,
            p_lm,
            e_r,
            e_theta,
            e_phi,
            cosine_theta,
            sine_theta,
            size_parameter,
            sph_h_derivative,
        ) = mutual_lookup(
            self.numerics.lmax,
            self.parameters.particles.position,
            sampling_points,
            self.parameters.k_medium,
            derivatives=True,
            parallel=False,
        )
        lookup_computation_time_stop = time()
        self.log.info(
            "Computing lookup tables took %f s",
            lookup_computation_time_stop - lookup_computation_time_start,
        )
        pi_lm, tau_lm = spherical_functions_trigon(
            self.numerics.lmax, cosine_theta, sine_theta
        )
        # print(sph_h.size)

        self.log.info("Computing field...")
        field_time_start = time()
        self.sampling_points = sampling_points
        if self.numerics.gpu:
            self.log.info("\t...using GPU")
            field_real = np.zeros(
                (self.parameters.k_medium.size, sampling_points.shape[0], 3),
                dtype=float,
            )
            field_imag = np.zeros_like(field_real)

            idx_device = cuda.to_device(self.idx_lookup)
            size_parameter_device = cuda.to_device(np.ascontiguousarray(size_parameter))
            sph_h_device = cuda.to_device(np.ascontiguousarray(sph_h))
            sph_h_derivative_device = cuda.to_device(
                np.ascontiguousarray(sph_h_derivative)
            )
            e_j_dm_phi_device = cuda.to_device(np.ascontiguousarray(e_j_dm_phi))
            p_lm_device = cuda.to_device(np.ascontiguousarray(p_lm))
            pi_lm_device = cuda.to_device(np.ascontiguousarray(pi_lm))
            tau_lm_device = cuda.to_device(np.ascontiguousarray(tau_lm))
            e_r_device = cuda.to_device(np.ascontiguousarray(e_r))
            e_theta_device = cuda.to_device(np.ascontiguousarray(e_theta))
            e_phi_device = cuda.to_device(np.ascontiguousarray(e_phi))
            sfc_device = cuda.to_device(
                np.ascontiguousarray(self.scattered_field_coefficients)
            )

            field_real_device = cuda.to_device(field_real)
            field_imag_device = cuda.to_device(field_imag)

            threads_per_block = (16, 16, 2)
            blocks_per_grid = (
                sampling_points.shape[0],
                sph_h.shape[1] * 2 * self.numerics.lmax * (self.numerics.lmax + 2),
                self.parameters.k_medium.size,
            )
            # blocks_per_grid = tuple(
            #     [
            #         ceil(blocks_per_grid[i] / threads_per_block[i])
            #         for i in range(len(threads_per_block))
            #     ]
            # )
            blocks_per_grid = tuple(
                ceil(blocks_per_grid[i] / threads_per_block[i])
                for i in range(len(threads_per_block))
            )

            compute_field_gpu[blocks_per_grid, threads_per_block](
                self.numerics.lmax,
                idx_device,
                size_parameter_device,
                sph_h_device,
                sph_h_derivative_device,
                e_j_dm_phi_device,
                p_lm_device,
                pi_lm_device,
                tau_lm_device,
                e_r_device,
                e_theta_device,
                e_phi_device,
                sfc_device,
                field_real_device,
                field_imag_device,
            )

            field_real = field_real_device.copy_to_host()
            field_imag = field_imag_device.copy_to_host()
            self.scattered_field = field_real + 1j * field_imag

        else:
            self.log.info("\t...using CPU")
            self.scattered_field = compute_field(
                self.numerics.lmax,
                self.idx_lookup,
                size_parameter,
                sph_h,
                sph_h_derivative,
                e_j_dm_phi,
                p_lm,
                pi_lm,
                tau_lm,
                e_r,
                e_theta,
                e_phi,
                scattered_field_coefficients=self.scattered_field_coefficients,
            )

        field_time_stop = time()
        self.log.info(
            f"\t Time taken for field calculation: {field_time_stop - field_time_start}"
        )



import logging
# import yasfpy.log as log

import numpy as np


class InitialField:
    """Represents an object with various parameters for a beam of light."""

    def __init__(
        self,
        beam_width,
        focal_point,
        field_type: str = "gaussian",
        amplitude: float = 1,
        polar_angle: float = 0,
        azimuthal_angle: float = 0,
        polarization: str = "TE",
    ):
        """Initializes an object with various parameters for a beam of light.

        Args:
            beam_width (float): The beam width parameter represents the width of the beam. It determines the spread of the beam
                and is typically measured as the full width at half maximum (FWHM) of the beam intensity
                distribution.
            focal_point (tuple): The focal point is the point in space where the beam is focused. It is a coordinate in
                three-dimensional space (x, y, z) that represents the position of the focal point.
            field_type (str, optional): The `field_type` parameter specifies the type of field for the beam. It can be set to
                "gaussian" or any other type of field that is supported by the code. Defaults to "gaussian".
            amplitude (float, optional): The amplitude parameter represents the maximum value or intensity of the beam. It determines
                the overall strength or power of the beam. Defaults to 1.
            polar_angle (float, optional): The `polar_angle` parameter represents the angle between the positive z-axis and the direction
                of propagation of the beam. It is measured in radians. Defaults to 0.
            azimuthal_angle (float, optional): The azimuthal angle is a measure of the angle between the projection of the vector onto the
                xy-plane and the positive x-axis. It determines the orientation of the beam in the xy-plane. Defaults to 0.
            polarization (str, optional): The "polarization" parameter determines the polarization of the beam. It can have two possible
                values: "TE" for transverse electric polarization and "TM" for transverse magnetic polarization. Defaults to "TE".
        """
        self.field_type = field_type
        self.amplitude = amplitude
        self.polar_angle = polar_angle
        self.azimuthal_angle = azimuthal_angle
        self.polarization = polarization
        self.beam_width = beam_width
        self.focal_point = focal_point

        # self.log = log.scattering_logger(__name__)
        self.log = logging.getLogger(self.__class__.__module__)
        self.__setup()

    def __set_pol_idx(self):
        """
        Sets the polarization index based on the polarization type.

        The polarization index is determined based on the value of the `polarization` attribute.
        If the `polarization` is "unp" or 0, the polarization index is set to 0.
        If the `polarization` is "te" or 1, the polarization index is set to 1.
        If the `polarization` is "tm" or 2, the polarization index is set to 2.
        If the `polarization` is not a valid value, the polarization index is set to 0 and a warning message is logged.
        """
        if (
            isinstance(self.polarization, str) and self.polarization.lower() == "unp"
        ) or (isinstance(self.polarization, int) and self.polarization == 0):
            # Unpolarized is also present in the MSTM output
            self.pol = 0
        elif (
            isinstance(self.polarization, str) and self.polarization.lower() == "te"
        ) or (isinstance(self.polarization, int) and self.polarization == 1):
            # Coresponds to the perpendicular value found in MSTM
            self.pol = 1
        elif (
            isinstance(self.polarization, str) and self.polarization.lower() == "tm"
        ) or (isinstance(self.polarization, int) and self.polarization == 2):
            # Coresponds to the parallel value found in MSTM
            self.pol = 2
        else:
            self.pol = 0
            self.log.warning(
                "%s is not a valid polarization type. Please use TE or TM. Reverting to unpolarized",
                self.polarization,
            )

    def __set_normal_incidence(self):
        """
        Sets the normal incidence flag based on the polar angle.

        This method checks the value of the polar angle and determines if it is close to zero.
        If the absolute value of the sine of the polar angle is less than 1e-5, the normal incidence flag is set to True.
        Otherwise, the normal incidence flag is set to False.
        """
        self.normal_incidence = np.abs(np.sin(self.polar_angle)) < 1e-5

    def __setup(self):
        """
        Performs the initial setup of the field.

        This method sets the polarization index and normal incidence for the field.
        """
        self.__set_pol_idx()
        self.__set_normal_incidence()



from yasfpy.particles import Particles
from yasfpy.initial_field import InitialField


import numpy as np


class Parameters:
    """
    The Parameters class represents the parameters for a simulation, including wavelength, refractive
    indices, scattering particles, and initial field, and provides methods for computing angular
    frequency and wave vectors.
    """

    def __init__(
        self,
        wavelength: np.ndarray,
        medium_refractive_index: np.ndarray,
        particles: Particles,
        initial_field: InitialField,
    ):
        """Initializes the class with the given parameters and sets up the necessary variables.

        Args:
            wavelength (np.array): An array that represents the wavelengths of the light being used.
                It contains the values of the wavelengths at which the simulation will be performed.
            medium_refractive_index (np.array): An array that represents the refractive index of the
                medium in which the particles are located. It contains the refractive index values at different
                wavelengths.
            particles (Particles): An instance of the "Particles" class. It represents the particles
                present in the medium.
            initial_field (InitialField): An object of the `InitialField` class. It represents the
                initial field configuration for the simulation.
        """
        self.wavelength = np.array(wavelength)
        self.medium_refractive_index = medium_refractive_index
        self.wavelengths_number = wavelength.size
        self.particles = particles
        self.initial_field = initial_field

        self.__setup()

    def __setup(self):
        """The function sets up the necessary computations for omega and ks."""
        self.__compute_omega()
        self.__compute_ks()

    def __compute_omega(self):
        """The function calculates the value of omega using the wavelength."""
        self.omega = 2 * np.pi / self.wavelength

    def __interpolate_refractive_index_from_table(self):
        """Interpolates the refractive index values from a table for different wavelengths.

        Returns:
            refractive_index_interpolated (np.array): An array that contains the interpolated refractive index values for the particles
                at different wavelengths.
        """
        refractive_index_interpolated = np.zeros(
            (self.particles.num_unique_refractive_indices, self.wavelength.size),
            dtype=complex,
        )
        for idx, data in enumerate(self.particles.refractive_index_table):
            table = data["ref_idx"].to_numpy().astype(float)
            n = np.interp(
                self.wavelength,
                table[:, 0],
                table[:, 1],
                left=table[0, 1],
                right=table[-1, 1],
            )
            k = np.interp(
                self.wavelength,
                table[:, 0],
                table[:, 2],
                left=table[0, 2],
                right=table[-1, 2],
            )
            refractive_index_interpolated[idx, :] = n + 1j * k
        return refractive_index_interpolated

    def __index_to_table(self):
        """
        Todo:
            do all the idx to value conversion here
        """
        pass

    def __compute_ks(self):
        """Computes the values of k_medium and k_particle based on the refractive index of the
        medium and particles.
        """
        self.k_medium = self.omega * self.medium_refractive_index
        if self.particles.refractive_index_table is None:
            self.ref_idx_table = None
            self.k_particle = np.outer(self.particles.refractive_index, self.omega)
        else:
            self.ref_idx_table = self.__interpolate_refractive_index_from_table()
            self.k_particle = (
                np.take(self.ref_idx_table, self.particles.refractive_index, axis=0)
                * np.array(self.omega)[np.newaxis, :]
            )

            unique_radius_index_pairs = np.zeros(
                (
                    self.particles.unique_radius_index_pairs.shape[0],
                    self.wavelength.size + 1,
                ),
                dtype=complex,
            )
            unique_radius_index_pairs[:, 0] = self.particles.unique_radius_index_pairs[
                :, 0
            ]
            unique_radius_index_pairs[:, 1:] = np.take(
                self.ref_idx_table,
                self.particles.unique_radius_index_pairs[:, 1].astype(int),
                axis=0,
            )

            self.particles.unique_radius_index_pairs = unique_radius_index_pairs



import logging
from typing import Union, Callable
from math import ceil
import pandas as pd
import numpy as np
from numba import cuda
from yasfpy.functions.misc import single_index2multi
from yasfpy.functions.spherical_functions_trigon import spherical_functions_trigon

from yasfpy.simulation import Simulation

from yasfpy.functions.cpu_numba import (
    compute_scattering_cross_section,
    compute_electric_field_angle_components,
    compute_polarization_components,
)
from yasfpy.functions.cuda_numba import (
    compute_scattering_cross_section_gpu,
    compute_electric_field_angle_components_gpu,
    compute_polarization_components_gpu,
)

# from yasfpy.functions.cpu_numba import compute_scattering_cross_section, compute_radial_independent_scattered_field, compute_electric_field_angle_components, compute_polarization_components
# from yasfpy.functions.cuda_numba import compute_scattering_cross_section_gpu, compute_radial_independent_scattered_field_gpu, compute_electric_field_angle_components_gpu, compute_polarization_components_gpu


class Optics:
    """The optics class handles the calculation of the optical scattering properties
    like scattering and extinction cross-sections.
    """

    def __init__(self, simulation: Simulation):
        """
        Initializes the Optics object.

        Args:
            simulation (Simulation): An instance of the Simulation class.

        """
        self.simulation = simulation

        self.c_ext = np.zeros_like(simulation.parameters.wavelength, dtype=complex)
        self.c_sca = np.zeros_like(simulation.parameters.wavelength, dtype=complex)

        self.log = logging.getLogger(self.__class__.__module__)

    def compute_cross_sections(self):
        """
        Compute the scattering and extinction cross sections.

        This method calculates the scattering and extinction cross sections for the simulation.
        It uses the initial field coefficients and scattered field coefficients to perform the calculations.
        """
        a = self.simulation.initial_field_coefficients
        f = self.simulation.scattered_field_coefficients

        self.c_ext = np.zeros(
            self.simulation.parameters.wavelengths_number, dtype=complex
        )
        self.c_ext -= (
            np.sum(np.conjugate(a) * f, axis=(0, 1))
            / np.power(self.simulation.parameters.k_medium, 2)
            * np.pi
        )

        lmax = self.simulation.numerics.lmax
        particle_number = self.simulation.parameters.particles.number
        wavelengths = self.simulation.parameters.k_medium.shape[0]
        translation_table = self.simulation.numerics.translation_ab5
        associated_legendre_lookup = self.simulation.plm
        spherical_bessel_lookup = self.simulation.sph_j
        e_j_dm_phi_loopup = self.simulation.e_j_dm_phi

        idx_lookup = self.simulation.idx_lookup

        if self.simulation.numerics.gpu:
            c_sca_real = np.zeros(wavelengths, dtype=float)
            c_sca_imag = np.zeros_like(c_sca_real)

            idx_device = cuda.to_device(idx_lookup)
            sfc_device = cuda.to_device(f)
            c_sca_real_device = cuda.to_device(c_sca_real)
            c_sca_imag_device = cuda.to_device(c_sca_imag)
            translation_device = cuda.to_device(translation_table)
            associated_legendre_device = cuda.to_device(associated_legendre_lookup)
            spherical_bessel_device = cuda.to_device(spherical_bessel_lookup)
            e_j_dm_phi_device = cuda.to_device(e_j_dm_phi_loopup)

            jmax = particle_number * 2 * lmax * (lmax + 2)
            threads_per_block = (16, 16, 2)
            blocks_per_grid_x = ceil(jmax / threads_per_block[0])
            blocks_per_grid_y = ceil(jmax / threads_per_block[1])
            blocks_per_grid_z = ceil(wavelengths / threads_per_block[2])
            blocks_per_grid = (blocks_per_grid_x, blocks_per_grid_y, blocks_per_grid_z)

            compute_scattering_cross_section_gpu[blocks_per_grid, threads_per_block](
                lmax,
                particle_number,
                idx_device,
                sfc_device,
                translation_device,
                associated_legendre_device,
                spherical_bessel_device,
                e_j_dm_phi_device,
                c_sca_real_device,
                c_sca_imag_device,
            )
            c_sca_real = c_sca_real_device.copy_to_host()
            c_sca_imag = c_sca_imag_device.copy_to_host()
            c_sca = c_sca_real + 1j * c_sca_imag

        else:
            # from numba_progress import ProgressBar
            # num_iterations = jmax * jmax * wavelengths
            # progress = ProgressBar(total=num_iterations)
            progress = None
            c_sca = compute_scattering_cross_section(
                lmax,
                particle_number,
                idx_lookup,
                f,
                translation_table,
                associated_legendre_lookup,
                spherical_bessel_lookup,
                e_j_dm_phi_loopup,
            )

        self.c_sca = (
            c_sca / np.power(np.abs(self.simulation.parameters.k_medium), 2) * np.pi
        )

        self.c_ext = np.real(self.c_ext)
        self.c_sca = np.real(self.c_sca)

        self.albedo = self.c_sca / self.c_ext

    def compute_phase_funcition(
        self,
        legendre_coefficients_number: int = 15,
        c_and_b: Union[bool, tuple] = False,
        check_phase_function: bool = False,
    ):
        """The function `compute_phase_function` calculates various polarization components and phase
        function coefficients for a given simulation.

        Args:
            legendre_coefficients_number (int, optional): The number of Legendre coefficients to compute
                for the phase function. These coefficients are used to approximate the phase function
                using Legendre polynomials. The higher the number of coefficients, the more accurate
                the approximation will be.
            c_and_b (Union[bool, tuple], optional): A boolean value or a tuple. If `True`, it indicates
                that the `c` and `b` bounds should be computed. If `False`, the function will return
                without computing the `c` and `b` bounds.
        """
        pilm, taulm = spherical_functions_trigon(
            self.simulation.numerics.lmax, self.simulation.numerics.polar_angles
        )

        if self.simulation.numerics.gpu:
            jmax = (
                self.simulation.parameters.particles.number
                * self.simulation.numerics.nmax
            )
            angles = self.simulation.numerics.azimuthal_angles.size
            wavelengths = self.simulation.parameters.k_medium.size
            e_field_theta_real = np.zeros(
                (
                    self.simulation.numerics.azimuthal_angles.size,
                    self.simulation.parameters.k_medium.size,
                ),
                dtype=float,
            )
            e_field_theta_imag = np.zeros_like(e_field_theta_real)
            e_field_phi_real = np.zeros_like(e_field_theta_real)
            e_field_phi_imag = np.zeros_like(e_field_theta_real)

            particles_position_device = cuda.to_device(
                self.simulation.parameters.particles.position
            )
            idx_device = cuda.to_device(self.simulation.idx_lookup)
            sfc_device = cuda.to_device(self.simulation.scattered_field_coefficients)
            k_medium_device = cuda.to_device(self.simulation.parameters.k_medium)
            azimuthal_angles_device = cuda.to_device(
                self.simulation.numerics.azimuthal_angles
            )
            e_r_device = cuda.to_device(self.simulation.numerics.e_r)
            pilm_device = cuda.to_device(pilm)
            taulm_device = cuda.to_device(taulm)
            e_field_theta_real_device = cuda.to_device(e_field_theta_real)
            e_field_theta_imag_device = cuda.to_device(e_field_theta_imag)
            e_field_phi_real_device = cuda.to_device(e_field_phi_real)
            e_field_phi_imag_device = cuda.to_device(e_field_phi_imag)

            sizes = (jmax, angles, wavelengths)
            threads_per_block = (16, 16, 2)
            # blocks_per_grid = tuple(
            #     [
            #         ceil(sizes[k] / threads_per_block[k])
            #         for k in range(len(threads_per_block))
            #     ]
            # )
            blocks_per_grid = tuple(
                ceil(sizes[k] / threads_per_block[k])
                for k in range(len(threads_per_block))
            )
            # blocks_per_grid = (
            #   ceil(jmax / threads_per_block[0]),
            #   ceil(angles / threads_per_block[1]),
            #   ceil(wavelengths / threads_per_block[2]))

            # print(f"{blocks_per_grid = }")

            compute_electric_field_angle_components_gpu[
                blocks_per_grid, threads_per_block
            ](
                self.simulation.numerics.lmax,
                particles_position_device,
                idx_device,
                sfc_device,
                k_medium_device,
                azimuthal_angles_device,
                e_r_device,
                pilm_device,
                taulm_device,
                e_field_theta_real_device,
                e_field_theta_imag_device,
                e_field_phi_real_device,
                e_field_phi_imag_device,
            )

            e_field_theta_real = e_field_theta_real_device.copy_to_host()
            e_field_theta_imag = e_field_theta_imag_device.copy_to_host()
            e_field_phi_real = e_field_phi_real_device.copy_to_host()
            e_field_phi_imag = e_field_phi_imag_device.copy_to_host()
            e_field_theta = e_field_theta_real + 1j * e_field_theta_imag
            e_field_phi = e_field_phi_real + 1j * e_field_phi_imag

            intensity = np.zeros_like(e_field_theta_real)
            dop = np.zeros_like(e_field_theta_real)
            dolp = np.zeros_like(e_field_theta_real)
            dolq = np.zeros_like(e_field_theta_real)
            dolu = np.zeros_like(e_field_theta_real)
            docp = np.zeros_like(e_field_theta_real)

            intensity_device = cuda.to_device(intensity)
            dop_device = cuda.to_device(dop)
            dolp_device = cuda.to_device(dolp)
            dolq_device = cuda.to_device(dolq)
            dolu_device = cuda.to_device(dolu)
            docp_device = cuda.to_device(docp)

            sizes = (angles, wavelengths)
            threads_per_block = (32, 32)
            # blocks_per_grid = tuple(
            #     [
            #         ceil(sizes[k] / threads_per_block[k])
            #         for k in range(len(threads_per_block))
            #     ]
            # )
            blocks_per_grid = tuple(
                ceil(sizes[k] / threads_per_block[k])
                for k in range(len(threads_per_block))
            )
            compute_polarization_components_gpu[blocks_per_grid, threads_per_block](
                self.simulation.parameters.k_medium.size,
                self.simulation.numerics.azimuthal_angles.size,
                e_field_theta_real_device,
                e_field_theta_imag_device,
                e_field_phi_real_device,
                e_field_phi_imag_device,
                intensity_device,
                dop_device,
                dolp_device,
                dolq_device,
                dolu_device,
                docp_device,
            )

            intensity = intensity_device.copy_to_host()
            dop = dop_device.copy_to_host()
            dolp = dolp_device.copy_to_host()
            dolq = dolq_device.copy_to_host()
            dolu = dolu_device.copy_to_host()
            docp = docp_device.copy_to_host()
        else:
            e_field_theta, e_field_phi = compute_electric_field_angle_components(
                self.simulation.numerics.lmax,
                self.simulation.parameters.particles.position,
                self.simulation.idx_lookup,
                self.simulation.scattered_field_coefficients,
                self.simulation.parameters.k_medium,
                self.simulation.numerics.azimuthal_angles,
                self.simulation.numerics.e_r,
                pilm,
                taulm,
            )

            intensity, dop, dolp, dolq, dolu, docp = compute_polarization_components(
                self.simulation.parameters.k_medium.size,
                self.simulation.numerics.azimuthal_angles.size,
                e_field_theta,
                e_field_phi,
            )

        self.scattering_angles = self.simulation.numerics.polar_angles
        k_medium = self.simulation.parameters.k_medium
        if type(self.simulation.parameters.k_medium) == pd.core.series.Series:
            k_medium = k_medium.to_numpy()
        self.phase_function_3d = (
            intensity
            * 4
            * np.pi
            / np.power(np.abs(k_medium), 2)
            / self.c_sca[np.newaxis, :]
        )
        if check_phase_function:
            res = self.__check_phase_function()
            assert (
                res is True
            ), "The phase function does have the desired precision. Please increase the amount of angles used."

        self.phase_function_legendre_coefficients = np.polynomial.legendre.legfit(
            np.cos(self.scattering_angles),
            self.phase_function_3d,
            legendre_coefficients_number,
        )

        self.degree_of_polarization_3d = dop
        self.degree_of_linear_polarization_3d = dolp
        self.degree_of_linear_polarization_q_3d = dolq
        self.degree_of_linear_polarization_u_3d = dolu
        self.degree_of_circular_polarization_3d = docp

        if (self.simulation.numerics.sampling_points_number is not None) and (
            self.simulation.numerics.sampling_points_number.size == 2
        ):
            self.phase_function = np.mean(
                np.reshape(
                    self.phase_function_3d,
                    np.append(
                        self.simulation.numerics.sampling_points_number,
                        self.simulation.parameters.k_medium.size,
                    ),
                ),
                axis=0,
            )

            self.degree_of_polarization = np.mean(
                np.reshape(
                    dop,
                    np.append(
                        self.simulation.numerics.sampling_points_number,
                        self.simulation.parameters.k_medium.size,
                    ),
                ),
                axis=0,
            )
            self.degree_of_linear_polarization = np.mean(
                np.reshape(
                    dolp,
                    np.append(
                        self.simulation.numerics.sampling_points_number,
                        self.simulation.parameters.k_medium.size,
                    ),
                ),
                axis=0,
            )
            self.degree_of_linear_polarization_q = np.mean(
                np.reshape(
                    dolq,
                    np.append(
                        self.simulation.numerics.sampling_points_number,
                        self.simulation.parameters.k_medium.size,
                    ),
                ),
                axis=0,
            )
            self.degree_of_linear_polarization_u = np.mean(
                np.reshape(
                    dolu,
                    np.append(
                        self.simulation.numerics.sampling_points_number,
                        self.simulation.parameters.k_medium.size,
                    ),
                ),
                axis=0,
            )
            self.degree_of_circular_polarization = np.mean(
                np.reshape(
                    docp,
                    np.append(
                        self.simulation.numerics.sampling_points_number,
                        self.simulation.parameters.k_medium.size,
                    ),
                ),
                axis=0,
            )

            self.scattering_angles = np.reshape(
                self.scattering_angles, self.simulation.numerics.sampling_points_number
            )
            self.scattering_angles = self.scattering_angles[0, :]
        else:
            self.phase_function = self.phase_function_3d

            self.degree_of_polarization = dop
            self.degree_of_linear_polarization = dolp
            self.degree_of_linear_polarization_q = dolq
            self.degree_of_linear_polarization_u = dolu
            self.degree_of_circular_polarization = docp

        self.c_and_b_bounds = c_and_b
        if isinstance(c_and_b, bool):
            if c_and_b:
                self.c_and_b_bounds = ([-1, 0], [1, 1])
            else:
                return

        self.__compute_c_and_b()

    def __check_phase_function(self, precision=1e-4) -> bool:
        integral = 0
        delta_phi = (
            len(self.simulation.numerics.azimuthal_angles)
            / self.simulation.numerics.sampling_points_number[0]
        )
        delta_theta = (
            len(self.simulation.numerics.polar_angles)
            / self.simulation.numerics.sampling_points_number[1]
        )
        for idx in self.simulation.numerics.azimuthal_angles:
            integral += (
                self.phase_function_3d[idx, :]
                * delta_phi
                * np.sin(self.simulation.numerics.azimuthal_angles[idx])
                * delta_theta
            )
        if 1 - integral / (4 * np.pi) > precision:
            return False
        return True

    def compute_asymmetry(self):
        """Computes the asymmetry parameter by numerical integration over the phase function.
        Therefore depends on the chosen sampling points. Accuracy depends on the number of
        sampling points. **ACCURACY CURRENTLY VERY MUCH IN QUESTION**
        """
        sum = 0
        delta_phi = (2 * np.pi) / self.simulation.numerics.sampling_points_number[0]
        delta_theta = np.pi / self.simulation.numerics.sampling_points_number[1]
        for idx in range(len(self.simulation.numerics.azimuthal_angles)):
            rn = np.cos(self.simulation.numerics.polar_angles[idx])
            a = (
                self.phase_function_3d[idx, :]
                * rn
                * delta_phi
                * np.sin(self.simulation.numerics.polar_angles[idx])
                * delta_theta
            )
            sum += a
        g = sum / (4 * np.pi)
        self.g = g

    def __compute_correction(self) -> float:
        """Naive implementation of a correction term to lessen the underestimation of a spheres surface
        via numerical integration. Does not adequately correct the result, but does improve it.

        Returns:
            correction_term (float): Computed error term
        """
        integral = 0
        delta_phi = (2 * np.pi) / self.simulation.numerics.sampling_points_number[0]
        delta_theta = np.pi / self.simulation.numerics.sampling_points_number[1]
        for idx in range(len(self.simulation.numerics.azimuthal_angles)):
            integral += (
                np.sin(self.simulation.numerics.polar_angles[idx])
                * delta_theta
                * delta_phi
            )
        error = 1 - integral / (4 * np.pi)
        return 12 * error / len(self.simulation.numerics.azimuthal_angles) ** 2

    def compute_asymmetry_corrected(self):
        """Computes asymmetry parameter via numerical integration over the phase function.
        Makes use of a correction term for the surface to compensate for underestimation
        of the spaces surface area. Therefore depends on the chosen sampling points.
        Accuracy depends on the number of sampling points.
        **ACCURACY CURRENTLY VERY MUCH IN QUESTION**
        """
        correction_term = self.__compute_correction()
        print(correction_term)
        sum = 0
        delta_phi = (2 * np.pi) / self.simulation.numerics.sampling_points_number[0]
        delta_theta = np.pi / self.simulation.numerics.sampling_points_number[1]
        for idx in range(len(self.simulation.numerics.azimuthal_angles)):
            rn = np.cos(self.simulation.numerics.polar_angles[idx])
            a = (
                self.phase_function_3d[idx, :]
                * rn
                * delta_phi
                * np.sin(self.simulation.numerics.polar_angles[idx])
                * delta_theta
            )
            correction = self.phase_function_3d[idx, :] * rn * correction_term
            sum += a
            sum += correction
        g = sum / (4 * np.pi)
        self.g_corr = g

    def compute_asymmetry_coeff(self):
        """WIP: Tried using the scattered field coefficients p & q to calculate the asymmetry parameter.
        Does **not** work. Might however be subject to further investigation,
        to achievce more reliable results.
        """
        A = self.simulation.scattered_field_coefficients
        a_i = []
        b_i = []
        for n in range(A.shape[1]):
            _, tau, _, _ = single_index2multi(n, self.simulation.numerics.lmax)
            if tau == 1:
                a_i.append(A[:, n, :])
            else:
                b_i.append(A[:, n, :])

        g = []
        for wv in range(self.simulation.parameters.wavelength.shape[0]):
            sum_term = 0
            for i_n in range(len(a_i) - 1):
                n = i_n + 1
                a_n = np.squeeze(a_i[i_n][:, wv])
                b_n = np.squeeze(b_i[i_n][:, wv])
                a_n2 = np.squeeze(a_i[i_n + 1][:, wv])
                b_n2 = np.squeeze(b_i[i_n + 1][:, wv])
                p1 = (
                    np.dot(a_n, np.conj(a_n2)) + np.dot(b_n, np.conj(b_n2))
                ).real  # *((n*(n+2))/(n+1))
                p2 = (np.dot(a_n, np.conj(b_n))).real  # *((2*n+1)/(n*(n+1)))
                sum_term += p1 + p2
            g_wv = (
                (4 * np.pi)
                / (
                    np.power(np.abs(self.simulation.parameters.k_particle[:, wv]), 2)
                    * self.c_sca[wv]
                )
            ) * sum_term
            g.append(g_wv)
        self.g = g
        return g

    @staticmethod
    def compute_double_henyey_greenstein(theta: np.ndarray, cb: np.ndarray):
        """
        Calculates the scattering phase function using the double Henyey-Greenstein model.

        Args:
            theta (np.ndarray): A numpy array representing the scattering angle. It contains the
                values at which you want to compute the double Henyey-Greenstein phase function.
            cb (np.ndarray): A numpy array that represents the coefficients of the double
                Henyey-Greenstein phase function. It should have a size of either 2 or 3.

        Returns:
            p (np.ndarray): Computed Henyey-Greenstein phase function.
        """
        cb = np.squeeze(cb)
        if cb.size < 2:
            cb = np.array([0, 0.5, 0.5])
        elif cb.size == 2:
            cb = np.append(cb, cb[1])
        elif cb.size > 3:
            cb = cb[:2]

        p1 = (1 - cb[1] ** 2) / np.power(
            1 - 2 * cb[1] * np.cos(theta) + cb[1] ** 2, 3 / 2
        )
        p2 = (1 - cb[2] ** 2) / np.power(
            1 + 2 * cb[2] * np.cos(theta) + cb[2] ** 2, 3 / 2
        )
        return (1 - cb[0]) / 2 * p1 + (1 + cb[0]) / 2 * p2

    def __compute_c_and_b(self):
        """The function computes the values of parameters 'b' and 'c' using the double Henyey-Greenstein
        phase function.
        """
        # double henyey greenstein
        if len(self.c_and_b_bounds[0]) not in [2, 3]:
            self.c_and_b_bounds = ([-1, 0], [1, 1])
            self.log.warning(
                "Number of parameters need to be 2 (b,c) or 3 (b1,b2,c). Reverting to two parameters (b,c) and setting the bounds to standard: b in [0, 1] and c in [-1, 1]"
            )

        from scipy.optimize import least_squares

        if len(self.c_and_b_bounds) == 2:
            bc0 = np.array([0, 0.5])
        else:
            bc0 = np.array([0, 0.5, 0.5])

        self.cb = np.empty((self.phase_function.shape[1], len(self.c_and_b_bounds)))
        for w in range(self.phase_function.shape[1]):
            # def dhg_optimization(bc):
            #     return (
            #         Optics.compute_double_henyey_greenstein(self.scattering_angles, bc)
            #         - self.phase_function[:, w]
            #     )

            # bc = least_squares(
            #     dhg_optimization, bc0, jac="2-point", bounds=self.c_and_b_bounds
            # )

            bc = least_squares(
                lambda bc: Optics.compute_double_henyey_greenstein(
                    self.scattering_angles, bc
                )
                - self.phase_function[:, w],
                bc0,
                jac="2-point",
                bounds=self.c_and_b_bounds,
            )
            self.cb[w, :] = bc.x

    def compute_phase_function(
        self,
        legendre_coefficients_number: int = 15,
        c_and_b: Union[bool, tuple] = False,
        check_phase_function: bool = False,
    ):
        """Generalized batching for phase function computation

        Args:
            legendre_coefficients_number (int, optional): The number of Legendre coefficients to compute
                for the phase function. These coefficients are used to approximate the phase function
                using Legendre polynomials. The higher the number of coefficients, the more accurate
                the approximation will be.
            c_and_b (Union[bool, tuple], optional): A boolean value or a tuple. If `True`, it indicates
                that the `c` and `b` bounds should be computed. If `False`, the function will return
                without computing the `c` and `b` bounds.
        """
        pilm, taulm = spherical_functions_trigon(
            self.simulation.numerics.lmax, self.simulation.numerics.polar_angles
        )

        if self.simulation.numerics.gpu:
            jmax = (
                self.simulation.parameters.particles.number
                * self.simulation.numerics.nmax
            )
            angles = self.simulation.numerics.azimuthal_angles.size
            wavelengths = self.simulation.parameters.k_medium.size
            e_field_theta_real = np.zeros(
                (
                    self.simulation.numerics.azimuthal_angles.size,
                    self.simulation.parameters.k_medium.size,
                ),
                dtype=float,
            )
            e_field_theta_imag = np.zeros_like(e_field_theta_real)
            e_field_phi_real = np.zeros_like(e_field_theta_real)
            e_field_phi_imag = np.zeros_like(e_field_theta_real)

            # stuff we need in full for every iteration
            device_data = [
                self.simulation.parameters.particles.position,
                self.simulation.idx_lookup,
                self.simulation.scattered_field_coefficients,
                self.simulation.parameters.k_medium,
            ]

            sizes = (jmax, angles, wavelengths)
            threads_per_block = (16, 16, 2)

            blocks_per_grid = tuple(
                ceil(sizes[k] / threads_per_block[k])
                for k in range(len(threads_per_block))
            )
            to_split = [
                np.ascontiguousarray(self.simulation.numerics.azimuthal_angles),
                np.ascontiguousarray(self.simulation.numerics.e_r),
                np.ascontiguousarray(pilm),
                np.ascontiguousarray(taulm),
                np.ascontiguousarray(e_field_theta_real),
                np.ascontiguousarray(e_field_theta_imag),
                np.ascontiguousarray(e_field_phi_real),
                np.ascontiguousarray(e_field_phi_imag),
            ]

            idx_to_split = self.simulation.numerics.azimuthal_angles.shape[0]

            idx_per_array = []
            for array in to_split:
                idx_per_array.append(
                    np.where(np.array(array.shape) == idx_to_split)[0][0]
                )

            threads_per_block = (1, 16 * 16, 2)
            external_args = [self.simulation.numerics.lmax]
            sizes_idx_split = 1
            res = self.__data_batching(
                external_args,
                device_data,
                to_split,
                sizes,
                sizes_idx_split,
                idx_to_split,
                idx_per_array,
                compute_electric_field_angle_components_gpu,
                threads_per_block,
            )
            e_field_theta_real = res[4]
            e_field_theta_imag = res[5]
            e_field_phi_real = res[6]
            e_field_phi_imag = res[7]

            e_field_theta = e_field_theta_real + 1j * e_field_theta_imag
            e_field_phi = e_field_phi_real + 1j * e_field_phi_imag

            print("...................................................")
            print("Done with e field calculations...")
            print("...................................................")
            # continue with next calculation

            intensity = np.zeros_like(e_field_theta_real)
            dop = np.zeros_like(e_field_theta_real)
            dolp = np.zeros_like(e_field_theta_real)
            dolq = np.zeros_like(e_field_theta_real)
            dolu = np.zeros_like(e_field_theta_real)
            docp = np.zeros_like(e_field_theta_real)

            to_split = [
                np.ascontiguousarray(e_field_theta_real),
                np.ascontiguousarray(e_field_theta_imag),
                np.ascontiguousarray(e_field_phi_real),
                np.ascontiguousarray(e_field_phi_imag),
                np.ascontiguousarray(intensity),
                np.ascontiguousarray(dop),
                np.ascontiguousarray(dolp),
                np.ascontiguousarray(dolq),
                np.ascontiguousarray(dolu),
                np.ascontiguousarray(docp),
            ]
            threads_per_block = (
                1024,
                1,
            )  # this allows ~65000 wavelengths, limits required batching

            idx_per_array = [0] * len(to_split)

            sizes = (angles, wavelengths)
            size_idx_split = 0

            external_args = [
                self.simulation.parameters.k_medium.size,
                self.simulation.numerics.azimuthal_angles.size,
            ]
            res = self.__data_batching(
                external_args,
                [],
                to_split,
                sizes,
                size_idx_split,
                idx_to_split,
                idx_per_array,
                compute_polarization_components_gpu,
                threads_per_block,
            )

            intensity = res[4]
            dop = res[5]
            dolp = res[6]
            dolq = res[7]
            dolu = res[8]
            docp = res[9]

        else:
            e_field_theta, e_field_phi = compute_electric_field_angle_components(
                self.simulation.numerics.lmax,
                self.simulation.parameters.particles.position,
                self.simulation.idx_lookup,
                self.simulation.scattered_field_coefficients,
                self.simulation.parameters.k_medium,
                self.simulation.numerics.azimuthal_angles,
                self.simulation.numerics.e_r,
                pilm,
                taulm,
            )

            intensity, dop, dolp, dolq, dolu, docp = compute_polarization_components(
                self.simulation.parameters.k_medium.size,
                self.simulation.numerics.azimuthal_angles.size,
                e_field_theta,
                e_field_phi,
            )

        self.scattering_angles = self.simulation.numerics.polar_angles
        k_medium = self.simulation.parameters.k_medium
        if type(self.simulation.parameters.k_medium) == pd.core.series.Series:
            k_medium = k_medium.to_numpy()
        self.phase_function_3d = (
            intensity
            * 4
            * np.pi
            / np.power(np.abs(k_medium), 2)
            / self.c_sca[np.newaxis, :]
        )
        if check_phase_function:
            res = self.__check_phase_function()
            assert (
                res is True
            ), "The phase function does have the desired precision. Please increase the amount of angles used."

        self.phase_function_legendre_coefficients = np.polynomial.legendre.legfit(
            np.cos(self.scattering_angles),
            self.phase_function_3d,
            legendre_coefficients_number,
        )

        self.degree_of_polarization_3d = dop
        self.degree_of_linear_polarization_3d = dolp
        self.degree_of_linear_polarization_q_3d = dolq
        self.degree_of_linear_polarization_u_3d = dolu
        self.degree_of_circular_polarization_3d = docp

        if (self.simulation.numerics.sampling_points_number is not None) and (
            self.simulation.numerics.sampling_points_number.size == 2
        ):
            self.phase_function = np.mean(
                np.reshape(
                    self.phase_function_3d,
                    np.append(
                        self.simulation.numerics.sampling_points_number,
                        self.simulation.parameters.k_medium.size,
                    ),
                ),
                axis=0,
            )

            self.degree_of_polarization = np.mean(
                np.reshape(
                    dop,
                    np.append(
                        self.simulation.numerics.sampling_points_number,
                        self.simulation.parameters.k_medium.size,
                    ),
                ),
                axis=0,
            )
            self.degree_of_linear_polarization = np.mean(
                np.reshape(
                    dolp,
                    np.append(
                        self.simulation.numerics.sampling_points_number,
                        self.simulation.parameters.k_medium.size,
                    ),
                ),
                axis=0,
            )
            self.degree_of_linear_polarization_q = np.mean(
                np.reshape(
                    dolq,
                    np.append(
                        self.simulation.numerics.sampling_points_number,
                        self.simulation.parameters.k_medium.size,
                    ),
                ),
                axis=0,
            )
            self.degree_of_linear_polarization_u = np.mean(
                np.reshape(
                    dolu,
                    np.append(
                        self.simulation.numerics.sampling_points_number,
                        self.simulation.parameters.k_medium.size,
                    ),
                ),
                axis=0,
            )
            self.degree_of_circular_polarization = np.mean(
                np.reshape(
                    docp,
                    np.append(
                        self.simulation.numerics.sampling_points_number,
                        self.simulation.parameters.k_medium.size,
                    ),
                ),
                axis=0,
            )

            self.scattering_angles = np.reshape(
                self.scattering_angles, self.simulation.numerics.sampling_points_number
            )
            self.scattering_angles = self.scattering_angles[0, :]
        else:
            self.phase_function = self.phase_function_3d

            self.degree_of_polarization = dop
            self.degree_of_linear_polarization = dolp
            self.degree_of_linear_polarization_q = dolq
            self.degree_of_linear_polarization_u = dolu
            self.degree_of_circular_polarization = docp

        self.c_and_b_bounds = c_and_b
        if isinstance(c_and_b, bool):
            if c_and_b:
                self.c_and_b_bounds = ([-1, 0], [1, 1])
            else:
                return

        self.__compute_c_and_b()

    def __data_batching(
        self,
        external_args: list,
        device_data: list[np.ndarray],
        to_split: list[np.ndarray],
        sizes: tuple,
        size_split_idx: int,
        idx_to_split: int,
        idx_per_array: list[int],
        cuda_kernel: Callable,
        threads_per_block: tuple,
    ):
        total_bytes = 0
        for array in to_split:
            total_bytes += array.size * array.itemsize
        # put device data onto array
        device_data2 = []
        for arr in device_data:
            device_data2.append(cuda.to_device(arr))

        start_idx = 0
        split_idx = 0
        done = False

        # build arrays for results
        # axis along which we split needs to be zero
        res = []
        for i, item in enumerate(to_split):
            needed_shape = list(item.shape)
            needed_shape[idx_per_array[i]] = 0
            needed_shape = tuple(needed_shape)
            res.append(np.empty(needed_shape, float))

        print(f"Need to process {total_bytes*1e-9} GB of data")

        while not done:
            split_idx = self.__compute_data_split(
                to_split,
                idx_list=idx_per_array,
                threads_per_block=threads_per_block[size_split_idx],
            )
            if split_idx < 1:
                break
            if split_idx + start_idx > idx_to_split:
                print("End, reset split_idx")
                split_idx = idx_to_split - start_idx

            split_data = []
            for i, item in enumerate(to_split):
                split_data.append(
                    np.take(
                        item,
                        range(start_idx, start_idx + split_idx),
                        axis=idx_per_array[i],
                    )
                )

            # check total size of data to be put on GPU
            used_bytes = 0
            for array in split_data:
                used_bytes += array.size * array.itemsize
            total_bytes -= used_bytes
            print(f"{total_bytes*1e-9} GB of data remaining")

            # modify number of blocks per grid needed for kernel with current split
            sizes2 = list(sizes)
            sizes2[size_split_idx] = len(range(start_idx, (start_idx + split_idx)))
            sizes = tuple(sizes2)
            blocks_per_grid = tuple(
                ceil(sizes[k] / threads_per_block[k])
                for k in range(len(threads_per_block))
            )

            # send data to device
            batched_device_data = []
            for i in range(len(to_split)):
                batched_device_data.append(cuda.to_device(split_data[i]))

            # call kernel
            arg_list = external_args + device_data2 + batched_device_data
            cuda_kernel[blocks_per_grid, threads_per_block](*arg_list)

            # receive batched data results
            for i, item in enumerate(batched_device_data):
                res[i] = np.append(
                    res[i],
                    np.array(item.copy_to_host()),
                    axis=idx_per_array[i],
                )

            # deallocate objects that use data on gpu so that cuda will deallocate memory
            del batched_device_data, split_data, arg_list

            # update start_idx
            start_idx += split_idx
            if start_idx >= idx_to_split:
                done = True

        return res

    def __compute_data_split(
        self, data: list[np.ndarray], idx_list: list, threads_per_block: int
    ) -> int:
        device = cuda.select_device(0)
        handle = cuda.cudadrv.devices.get_context()
        mem_info = cuda.cudadrv.driver.Context(device, handle).get_memory_info()
        buffer = 0.05 * mem_info.total  # buffer to accomodate for varying GPU mem usage
        free_bytes = mem_info.free - buffer
        total_data_bytes = 0
        for array in data:
            total_data_bytes += array.size * array.itemsize

        print("---------------------------------------------------")
        idx = data[0].shape[idx_list[0]]
        num = idx
        while total_data_bytes > free_bytes:
            new_data_bytes = 0
            num -= 1000
            for i, item in enumerate(data):
                temp_shape = item.shape
                temp_size = 1
                for s in temp_shape:
                    if s == idx:
                        temp_size *= num
                    else:
                        temp_size *= s
                new_data_bytes += temp_size * item.itemsize
            total_data_bytes = new_data_bytes

        print(f"{free_bytes*1e-9} GB of data available on GPU")
        print(f"Unused GPU memory: {(free_bytes-total_data_bytes)*1e-6} MB")

        if num // threads_per_block > 2**16 - 1:
            num = (2**16 - 1) * threads_per_block
            print(
                "Need to limit number of blocks due to limited number of blocks per grid!"
            )

        print("---------------------------------------------------")
        return num



import logging
# import yasfpy.log as log

import numpy as np
from scipy.sparse.linalg import LinearOperator, gmres, lgmres, bicgstab


class Solver:
    """
    The Solver class provides a generic interface for solving linear systems of equations using
    different iterative solvers such as GMRES, BiCGSTAB, and LGMRES, and the GMResCounter class is used
    to count the number of iterations and display the residual or current iterate during the GMRES
    solver.
    """

    def __init__(
        self,
        solver_type: str = "gmres",
        tolerance: float = 1e-4,
        max_iter: float = 1e4,
        restart: float = 1e2,
    ):
        """Initializes a solver object with specified parameters and creates a logger object.

        Args:
            solver_type (str, optional): The type of solver to be used. Defaults to "gmres".
            tolerance (float): The desired accuracy of the solver.
            max_iter (int): The maximum number of iterations that the solver will perform.
            restart (int): The number of iterations after which the solver will restart.
        """
        self.type = solver_type.lower()
        self.tolerance = tolerance
        self.max_iter = int(max_iter)
        self.restart = int(restart)

        # self.log = log.scattering_logger(__name__)
        self.log = logging.getLogger(self.__class__.__module__)

    def run(self, a: LinearOperator, b: np.ndarray, x0: np.ndarray = None):
        """
        Runs the solver on the given linear system of equations.

        Args:
            a (LinearOperator): The linear operator representing the system matrix.
            b (np.ndarray): The right-hand side vector.
            x0 (np.ndarray, optional): The initial guess for the solution. If not provided, a copy of b will be used.

        Returns:
            value (np.ndarray): The solution to the linear system of equations.
            err_code (int): The error code indicating the convergence status of the solver.

        """
        if x0 is None:
            x0 = np.copy(b)

        if np.any(np.isnan(b)):
            print(b)

        if self.type == "bicgstab":
            # Add your code here for the bicgstab solver
            pass
            counter = GMResCounter(callback_type="x")
            value, err_code = bicgstab(
                a,
                b,
                x0,
                tol=self.tolerance,
                atol=0,
                maxiter=self.max_iter,
                callback=counter,
            )
        elif self.type == "gmres":
            counter = GMResCounter(callback_type="pr_norm")
            value, err_code = gmres(
                a,
                b,
                x0,
                restart=self.restart,
                tol=self.tolerance,
                atol=self.tolerance**2,
                maxiter=self.max_iter,
                callback=counter,
                callback_type="pr_norm",
            )
        elif self.type == "lgmres":
            counter = GMResCounter(callback_type="x")
            value, err_code = lgmres(
                a,
                b,
                x0,
                rtol=self.tolerance,
                atol=self.tolerance**2,
                maxiter=self.max_iter,
                callback=counter,
            )
        else:
            self.log.error("Please specify a valid solver type")
            exit(1)

        return value, err_code


class GMResCounter(object):
    """
    The GMResCounter class is a helper class that counts the number of iterations and displays the
    residual or current iterate during the GMRES solver.
    """

    def __init__(self, disp: bool = False, callback_type: str = "pr_norm"):
        """Initializes an object with optional display and callback type parameters.

        Args:
            disp (bool, optional): A boolean flag that determines whether or not to display the progress
                of the algorithm. If `disp` is set to `True`, the algorithm will display the progress.
                If `disp` is set to `False`, the algorithm will not display the progress.
            callback_type (str, optional): The type of callback to be used. It can have two possible values.

        """
        # self.log = log.scattering_logger(__name__)
        self.log = logging.getLogger(self.__class__.__module__)
        self._disp = disp
        self.niter = 0
        if callback_type == "pr_norm":
            # self.header = "% 10s \t % 15s" % ("Iteration", "Residual")
            self.header = " Iteration \t        Residual"
        elif callback_type == "x":
            # self.header = "% 10s \t %s" % ("Iteration", "Current Iterate")
            self.header = " Iteration \t Current Iterate"

    def __call__(self, rk=None):
        """The function increments a counter, formats a message based on the input, logs the header and
        message, and prints the header and message if the `_disp` flag is True.

        Args:
            rk (Union[np.ndarray, float]): The parameter `rk` can be either a float or a numpy array.

        """
        self.niter += 1
        if isinstance(rk, float):
            # msg = "% 10i \t % 15.5f" % (self.niter, rk)
            msg = f"{self.niter:10} \t {rk:15.5f}"
        elif isinstance(rk, np.ndarray):
            # msg = "% 10i \t " % self.niter + np.array2string(rk)
            msg = f"{self.niter:10} \t {np.array2string(rk)}"

        self.log.debug(self.header)
        self.log.debug(msg)
        if self._disp:
            print(self.header)
            print(msg)



import re
import io
import pandas as pd
import yaml

# import urllib.request
from urllib.parse import unquote
import requests as req

import numpy as np


def material_handler(links):
    """
    Handles the processing of material data from various sources.

    Args:
        links (str or list): The link(s) to the material data source(s).

    Returns:
        (dict): A dictionary containing the processed material data and information.

    """
    if not isinstance(links, list):
        links = [links]

    data = dict(ref_idx=pd.DataFrame(columns=["wavelength", "n", "k"]), material=None)
    for link in links:
        link = link.strip()
        if link[:4] == "http":
            if "refractiveindex.info" in link:
                df, material = handle_refractiveindex_info(link)
                data["ref_idx"] = pd.concat([data["ref_idx"], df])
                data["material"] = material
            elif "http://eodg.atm.ox.ac.uk" in link:
                df, material = handle_eodg(link)
                data["ref_idx"] = pd.concat([data["ref_idx"], df])
                data["material"] = material
            else:
                print("No matching handler found for url")
        else:
            if ".csv" in link:
                df, material = handle_csv(link)
                data["ref_idx"] = pd.concat([data["ref_idx"], df])
                data["material"] = material
            else:
                print("No matching handler found for file type")
    data["ref_idx"] = data["ref_idx"].sort_values(by=["wavelength"])
    return data


def handle_refractiveindex_info(url):
    """
    Retrieves refractive index data from a given URL and processes it.

    Args:
        url (str): The URL to retrieve the refractive index data from.

    Returns:
        (tuple): A tuple containing the processed data as a pandas DataFrame and the material name.

    Raises:
        Exception: If the data retrieval fails.

    """
    url_split = url.replace("=", "/").split("/")
    material = unquote(url_split[-2])

    if np.any([("data_csv" in part) or ("data_txt" in part) for part in url_split]):
        print("Please use the [Full database record] option for refractiveindex.info!")
        print("Reverting url:")
        print(f" from: {url}")
        url_split[3] = "database"
        url = "/".join(url_split)
        print(f" to:   {url}")

    # req = urllib.request.Request(url)
    # with urllib.request.urlopen(req) as resp:
    resp = req.get(url)
    if resp.status_code >= 400:
        raise Exception(f"Failed to retrieve data from {url}")
    # data = resp.read()
    # data = data.decode("utf-8")
    data = resp.text
    data_yml = yaml.safe_load(data)
    header_yml = ["wavelength", "n", "k"]
    data = pd.DataFrame(columns=["wavelength", "n", "k"])
    for line in data_yml["DATA"]:
        df = None
        if "tabulated" in line["type"].lower():
            # elif line['type'].lower()[-2:] == ' n':
            #   header_yml=['wavelength', 'n']
            if line["type"].lower()[-2:] == " k":
                header_yml = ["wavelength", "k", "n"]
            df = pd.read_csv(
                io.StringIO(line["data"]),
                delim_whitespace=True,
                header=None,
                names=header_yml,
            )
        elif "formula" in line["type"].lower():
            if line["type"].lower() == "formula 1":
                wavelengths = [float(c) for c in line["wavelength_range"].split()]
                wavelengths = np.arange(wavelengths[0], wavelengths[1], 0.1)
                coefficients = np.array(
                    [float(c) for c in line["coefficients"].split()]
                )
                ref_idx = lambda x: np.sqrt(
                    1
                    + np.sum(
                        [
                            coefficients[i] * x**2 / (x**2 - coefficients[i + 1] ** 2)
                            for i in range(1, len(coefficients), 2)
                        ],
                        axis=0,
                    )
                )
                df = pd.DataFrame(columns=["wavelength", "n", "k"])
                df["wavelength"] = wavelengths
                df["n"] = ref_idx(wavelengths)

        if df is not None:
            df = df.fillna(0)
            data = pd.concat([data, df])

    return data, material


def handle_eodg(url):
    """
    Handles the EODG material data retrieval from a given URL.

    Args:
        url (str): The URL of the EODG material data.

    Returns:
        (tuple): A tuple containing the retrieved material data and the material name.

    Raises:
        Exception: If the data retrieval from the URL fails.
    """
    url_split = url.split("/")
    material = unquote(url_split[6])

    resp = req.get(url)
    if resp.status_code >= 400:
        raise Exception(f"Failed to retrieve data from {url}")
    data = resp.text
    data_format = [
        s.lower() for s in re.search(r"#FORMAT=(.*)\n", data).group(1).split()
    ]
    header_yml = ["wavelength", "n", "k"]
    if "n" not in data_format:
        header_yml = ["wavelength", "k", "n"]

    data = re.sub(r"^#.*\n", "", data, flags=re.MULTILINE)
    data = pd.read_csv(
        io.StringIO(data), delim_whitespace=True, header=None, names=header_yml
    )
    data = data.fillna(0)
    if "wavn" in data_format:
        data["wavelength"] = 1e4 / data["wavelength"]
        data = data.iloc[::-1]

    return data, material


def handle_csv(path):
    """
    Read a CSV file containing material data and extract the material name and data.

    Args:
        path (str): The path to the CSV file.

    Returns:
        (tuple): A tuple containing the material data as a pandas DataFrame and the material name.
    """
    name = re.split(r"\._-", path)
    material = unquote(name[0])
    data = pd.read_csv(
        path, delim_whitespace=False, header=0, names=["wavelength", "n", "k"]
    )
    return data, material



from yasfpy.particles import Particles
from yasfpy.functions.legendre_normalized_trigon import legendre_normalized_trigon

import numpy as np
from scipy.special import spherical_jn
from scipy.special import hankel1, lpmv


def jmult_max(num_part, lmax):
    """
    Calculate the maximum value of jmult.

    Parameters:
    num_part (int): The number of particles.
    lmax (int): The maximum value of l.

    Returns:
        (int): The maximum value of jmult.
    """
    return num_part * 2 * lmax * (lmax + 2)


def multi2single_index(j_s, tau, l, m, lmax):
    """
    Converts the multi-index (j_s, tau, l, m) to a single index.

    Args:
        j_s (int): Particle index.
        tau (int): Polarization value (1 or 2).
        l (int): The value of l (between 0 and lmax).
        m (int): The value of m (between -l and l).
        lmax (int): Cutoff value for the field expansion.

    Returns:
        (int): The single index corresponding to the multi-index (j_s, tau, l, m).
    """
    return (
        j_s * 2 * lmax * (lmax + 2)
        + (tau - 1) * lmax * (lmax + 2)
        + (l - 1) * (l + 1)
        + m
        + l
    )


def single_index2multi(idx, lmax):
    """
    Convert a single index to multi-indices (j_s, tau, l, m) for spherical harmonics.

    Args:
        idx (int): The single index.
        lmax (int): Cutoff value for the field expansion.

    Returns:
        j_s (int): Particle index.
        tau (int): Polarization value (1 or 2).
        l (float): The value of l (between 0 and lmax).
        m (int): The value of m (between -l and l).
    """
    j_s = idx // (2 * lmax * (lmax + 2))
    idx_new = idx % (2 * lmax * (lmax + 2))
    tau = idx_new // (lmax * (lmax + 2)) + 1
    idx_new = idx_new % (lmax * (lmax + 2))
    l = np.floor(np.sqrt(idx_new + 1))
    m = idx_new - (l * l + l - 1)
    return j_s, tau, l, m


def transformation_coefficients(pilm, taulm, tau, l, m, pol, dagger: bool = False):
    """
    Calculate the transformation coefficients for spherical harmonics.

    Args:
        pilm (ndarray): Array of spherical harmonics.
        taulm (ndarray): Array of spherical harmonics.
        tau (int): Polarization state.
        l (int): Degree of the spherical harmonics.
        m (int): Order of the spherical harmonics.
        pol (int): Polarization state.
        dagger (bool, optional): Whether to apply the dagger operation. Defaults to False.

    Returns:
        (float): The transformation coefficient.

    """
    ifac = 1j
    if dagger:
        ifac *= -1

    # Polarized light
    if np.any(np.equal(pol, [1, 2])):
        if tau == pol:
            spher_fun = taulm[l, np.abs(m)]
        else:
            spher_fun = m * pilm[l, np.abs(m)]

        return (
            -1
            / np.power(ifac, l + 1)
            / np.sqrt(2 * l * (l + 1))
            * (ifac * (pol == 1) + (pol == 2))
            * spher_fun
        )

    # Unpolarized light
    return (
        -1
        / np.power(ifac, l + 1)
        / np.sqrt(2 * l * (l + 1))
        * (ifac + 1)
        * (taulm[l, np.abs(m)] + m * pilm[l, np.abs(m)])
        / 2
    )


def mutual_lookup(
    lmax: int,
    positions_1: np.ndarray,
    positions_2: np.ndarray,
    refractive_index: np.ndarray,
    derivatives: bool = False,
    parallel: bool = False,
):
    """
    Calculate mutual lookup tables for scattering calculations.

    Args:
        lmax (int): The maximum degree of the spherical harmonics expansion.
        positions_1 (np.ndarray): The positions of the first set of particles.
        positions_2 (np.ndarray): The positions of the second set of particles.
        refractive_index (np.ndarray): The refractive indices of the particles.
        derivatives (bool, optional): Whether to calculate the derivatives of the lookup tables. Defaults to False.
        parallel (bool, optional): Whether to use parallel computation. Defaults to False.

    Returns:
        spherical_bessel (np.ndarray): The spherical Bessel functions.
        spherical_hankel (np.ndarray): The spherical Hankel functions.
        e_j_dm_phi (np.ndarray): The exponential term in the scattering calculation.
        p_lm (np.ndarray): The normalized Legendre polynomials.
        e_r (np.ndarray): The unit vectors in the radial direction.
        e_theta (np.ndarray): The unit vectors in the polar direction.
        e_phi (np.ndarray): The unit vectors in the azimuthal direction.
        cosine_theta (np.ndarray): The cosine of the polar angle.
        sine_theta (np.ndarray): The sine of the polar angle.
        size_parameter (np.ndarray): The size parameter of the particles.
        spherical_hankel_derivative (np.ndarray): The derivative of the spherical Hankel functions.
    """
    differences = positions_1[:, np.newaxis, :] - positions_2[np.newaxis, :, :]
    distances = np.sqrt(np.sum(differences**2, axis=2))
    distances = distances[:, :, np.newaxis]
    e_r = np.divide(
        differences, distances, out=np.zeros_like(differences), where=distances != 0
    )
    cosine_theta = e_r[:, :, 2]
    # cosine_theta = np.divide(
    #   differences[:, :, 2],
    #   distances,
    #   out = np.zeros_like(distances),
    #   where = distances != 0
    # )
    # correct possible rounding errors
    cosine_theta[cosine_theta < -1] = -1
    cosine_theta[cosine_theta > 1] = 1
    sine_theta = np.sqrt(1 - cosine_theta**2)
    phi = np.arctan2(differences[:, :, 1], differences[:, :, 0])
    e_theta = np.stack(
        [cosine_theta * np.cos(phi), cosine_theta * np.sin(phi), -sine_theta], axis=-1
    )
    e_phi = np.stack([-np.sin(phi), np.cos(phi), np.zeros_like(phi)], axis=-1)

    size_parameter = distances * np.array(refractive_index)[np.newaxis, np.newaxis, :]

    if parallel:
        from yasfpy.functions.cpu_numba import compute_lookup_tables

        spherical_bessel, spherical_hankel, e_j_dm_phi, p_lm = compute_lookup_tables(
            lmax, size_parameter, phi, cosine_theta
        )
    else:
        p_range = np.arange(2 * lmax + 1)
        p_range = p_range[:, np.newaxis, np.newaxis, np.newaxis]
        size_parameter_extended = size_parameter[np.newaxis, :, :, :]
        spherical_hankel = np.sqrt(
            np.divide(
                np.pi / 2,
                size_parameter_extended,
                out=np.zeros_like(size_parameter_extended),
                where=size_parameter_extended != 0,
            )
        ) * hankel1(p_range + 1 / 2, size_parameter_extended)
        spherical_bessel = spherical_jn(p_range, size_parameter_extended)

        if derivatives:
            spherical_hankel_lower = np.sqrt(
                np.divide(
                    np.pi / 2,
                    size_parameter_extended,
                    out=np.zeros_like(size_parameter_extended),
                    where=size_parameter_extended != 0,
                )
            ) * hankel1(-1 / 2, size_parameter_extended)
            spherical_hankel_lower = np.vstack(
                (spherical_hankel_lower, spherical_hankel[:-1, :, :, :])
            )
            spherical_hankel_derivative = (
                size_parameter_extended * spherical_hankel_lower
                - p_range * spherical_hankel
            )

            # p_range = np.arange(2 * lmax + 2) - 1
            # p_range = p_range[:, np.newaxis, np.newaxis, np.newaxis]
            # spherical_hankel = np.sqrt(np.divide(np.pi / 2, size_parameter_extended, out = np.zeros_like(size_parameter_extended), where = size_parameter_extended != 0)) * hankel1(p_range + 1/2, size_parameter_extended)
            # spherical_hankel_derivative = size_parameter_extended * spherical_hankel[:-1, :, :, :] - p_range[1:, :, :, :] * spherical_hankel[1:, :, :, :]

            p_lm = legendre_normalized_trigon(lmax, cosine_theta, sine_theta)
        else:
            spherical_hankel_derivative = None

            p_lm = np.zeros(
                (lmax + 1) * (2 * lmax + 1) * np.prod(size_parameter.shape[:2])
            ).reshape(((lmax + 1) * (2 * lmax + 1),) + size_parameter.shape[:2])
            for p in range(2 * lmax + 1):
                for absdm in range(p + 1):
                    cml = np.sqrt(
                        (2 * p + 1)
                        / 2
                        * np.prod(1 / np.arange(p - absdm + 1, p + absdm + 1))
                    )
                    # if np.isnan(cml):
                    #     print(p)
                    #     print(absdm)
                    p_lm[p * (p + 1) // 2 + absdm, :, :] = (
                        cml * np.power(-1.0, absdm) * lpmv(absdm, p, cosine_theta)
                    )

        phi = phi[np.newaxis, :, :]
        p_range = np.arange(-2 * lmax, 2 * lmax + 1)
        p_range = p_range[:, np.newaxis, np.newaxis]
        e_j_dm_phi = np.exp(1j * p_range * phi)

    return (
        spherical_bessel,
        spherical_hankel,
        e_j_dm_phi,
        p_lm,
        e_r,
        e_theta,
        e_phi,
        cosine_theta,
        sine_theta,
        size_parameter,
        spherical_hankel_derivative,
    )


def interpolate_refractive_index_from_table(
    wavelengths: np.ndarray, materials: list, species_idx: np.ndarray
) -> np.ndarray:
    """Interpolates the refractive index values from a table for different wavelengths.

    Returns:
        refractive_index_interpolated (np.array): An array that contains the interpolated refractive index values for the particles
            at different wavelengths.
    """

    refractive_index_table = Particles.generate_refractive_index_table(materials)

    unique_refractive_indices, _ = np.unique(species_idx, return_inverse=True, axis=0)
    num_unique_refractive_indices = unique_refractive_indices.shape[0]

    refractive_index_interpolated = np.zeros(
        (num_unique_refractive_indices, wavelengths.size),
        dtype=complex,
    )
    for idx, data in enumerate(refractive_index_table):
        table = data["ref_idx"].to_numpy().astype(float)
        n = np.interp(
            wavelengths,
            table[:, 0],
            table[:, 1],
            left=table[0, 1],
            right=table[-1, 1],
        )
        k = np.interp(
            wavelengths,
            table[:, 0],
            table[:, 2],
            left=table[0, 2],
            right=table[-1, 2],
        )
        refractive_index_interpolated[idx, :] = n + 1j * k
    return refractive_index_interpolated



from cmath import exp, sqrt
import numpy as np
from numba import cuda
from yasfpy.computers import Computer


class ComputerCUDA(Computer):
    # TODO: Implement data batching for GPUs with smaller memory
    @cuda.jit(fastmath=True)
    def particle_interaction(
        lmax: int,
        particle_number: int,
        idx: np.ndarray,
        x: np.ndarray,
        wx_real: np.ndarray,
        wx_imag: np.ndarray,
        translation_table: np.ndarray,
        plm: np.ndarray,
        sph_h: np.ndarray,
        e_j_dm_phi,
    ):
        """
        Perform particle interaction calculations on the GPU.

        Args:
            lmax (int): Maximum angular momentum quantum number.
            particle_number (int): Number of particles.
            idx (np.ndarray): Array containing particle indices.
            x (np.ndarray): Array of particle positions.
            wx_real (np.ndarray): Array to store the real part of the result.
            wx_imag (np.ndarray): Array to store the imaginary part of the result.
            translation_table (np.ndarray): Array containing translation table.
            plm (np.ndarray): Array containing associated Legendre polynomials.
            sph_h (np.ndarray): Array containing spherical harmonics.
            e_j_dm_phi (np.ndarray): Additional parameter for the calculation.

        Todo:
            Implement data batching for GPUs with smaller memory
        """
        j1, j2, w = cuda.grid(3)

        jmax = particle_number * 2 * lmax * (lmax + 2)
        channels = sph_h.shape[-1]

        if (j1 >= jmax) or (j2 >= jmax) or (w >= channels):
            return

        s1, n1, tau1, l1, m1 = idx[j1, :]
        s2, n2, tau2, l2, m2 = idx[j2, :]

        if s1 == s2:
            return

        delta_tau = abs(tau1 - tau2)
        delta_l = abs(l1 - l2)
        delta_m = abs(m1 - m2)

        p_dependent = complex(0)
        for p in range(max(delta_m, delta_l + delta_tau), l1 + l2 + 1):
            p_dependent += (
                translation_table[n2, n1, p]
                * plm[p * (p + 1) // 2 + delta_m, s1, s2]
                * sph_h[p, s1, s2, w]
            )
        p_dependent *= e_j_dm_phi[m2 - m1 + 2 * lmax, s1, s2] * x[j2]

        # atomic.add performs the += operation in sync
        cuda.atomic.add(wx_real, (j1, w), p_dependent.real)
        cuda.atomic.add(wx_imag, (j1, w), p_dependent.imag)

    @cuda.jit(fastmath=True)
    def compute_scattering_cross_section(
        lmax: int,
        particle_number: int,
        idx: np.ndarray,
        sfc: np.ndarray,
        translation_table: np.ndarray,
        plm: np.ndarray,
        sph_h: np.ndarray,
        e_j_dm_phi: np.ndarray,
        c_sca_real: np.ndarray,
        c_sca_imag: np.ndarray,
    ):
        """
        Compute the scattering cross section on the GPU using CUDA.

        Args:
            lmax (int): The maximum degree of the spherical harmonics expansion.
            particle_number (int): The number of particles.
            idx (np.ndarray): The index array.
            sfc (np.ndarray): The scattering form factor array.
            translation_table (np.ndarray): The translation table array.
            plm (np.ndarray): The associated Legendre polynomials array.
            sph_h (np.ndarray): The spherical harmonics array.
            e_j_dm_phi (np.ndarray): The phase factor array.
            c_sca_real (np.ndarray): The real part of the scattering cross section array.
            c_sca_imag (np.ndarray): The imaginary part of the scattering cross section array.
        """
        jmax = particle_number * 2 * lmax * (lmax + 2)
        channels = sph_h.shape[-1]

        j1, j2, w = cuda.grid(3)

        if (j1 >= jmax) or (j2 >= jmax) or (w >= channels):
            return

        s1, n1, _, _, m1 = idx[j1, :]
        s2, n2, _, _, m2 = idx[j2, :]

        delta_m = abs(m1 - m2)

        p_dependent = complex(0)
        for p in range(delta_m, 2 * lmax + 1):
            p_dependent += (
                translation_table[n2, n1, p]
                * plm[p * (p + 1) // 2 + delta_m, s1, s2]
                * sph_h[p, s1, s2, w]
            )
        p_dependent *= (
            sfc[s1, n1, w].conjugate()
            * e_j_dm_phi[m2 - m1 + 2 * lmax, s1, s2]
            * sfc[s2, n2, w]
        )

        # atomic.add performs the += operation in sync
        cuda.atomic.add(c_sca_real, w, p_dependent.real)
        cuda.atomic.add(c_sca_imag, w, p_dependent.imag)

    @cuda.jit(fastmath=True)
    def compute_radial_independent_scattered_field(
        lmax: int,
        particles_position: np.ndarray,
        idx: np.ndarray,
        sfc: np.ndarray,
        k_medium: np.ndarray,
        azimuthal_angles: np.ndarray,
        e_r: np.ndarray,
        e_phi: np.ndarray,
        e_theta: np.ndarray,
        pilm: np.ndarray,
        taulm: np.ndarray,
        e_1_sca_real: np.ndarray,
        e_1_sca_imag: np.ndarray,
    ):
        """
        Compute the radial independent scattered field using GPU acceleration.

        Args:
            lmax (int): The maximum degree of the spherical harmonics expansion.
            particles_position (np.ndarray): Array of particle positions.
            idx (np.ndarray): Array of indices for particle properties.
            sfc (np.ndarray): Array of scattering form factors.
            k_medium (np.ndarray): Array of wave numbers in the medium.
            azimuthal_angles (np.ndarray): Array of azimuthal angles.
            e_r (np.ndarray): Array of radial electric field components.
            e_phi (np.ndarray): Array of azimuthal electric field components.
            e_theta (np.ndarray): Array of polar electric field components.
            pilm (np.ndarray): Array of associated Legendre polynomials.
            taulm (np.ndarray): Array of tau coefficients.
            e_1_sca_real (np.ndarray): Array of real parts of the scattered electric field.
            e_1_sca_imag (np.ndarray): Array of imaginary parts of the scattered electric field.
        """
        j_idx, a_idx, w_idx = cuda.grid(3)

        jmax = particles_position.shape[0] * 2 * lmax * (lmax + 2)

        if (
            (j_idx >= jmax)
            or (a_idx >= azimuthal_angles.size)
            or (w_idx >= k_medium.size)
        ):
            return

        s, n, tau, l, m = idx[j_idx, :]

        # Temporary variable
        # If tau = 1 -> 1j**(tau-1) = 1, if tau = 2 -> 1j**(tau-1) = 1j
        # 1j**(-l-1) = (-1j)**(l+1) => both lead to the coefficient 1j**(tau-l-2)
        # k * <particle_position, e_r> is the phase shift due to the distance and relative position
        t = (
            1j ** (tau - l - 2)
            * sfc[s, n, w_idx]
            / sqrt(2 * l * (l + 1))
            * exp(
                1j
                * (
                    m * azimuthal_angles[a_idx]
                    - k_medium[w_idx]
                    * (
                        particles_position[s, 0] * e_r[a_idx, 0]
                        + particles_position[s, 1] * e_r[a_idx, 1]
                        + particles_position[s, 2] * e_r[a_idx, 2]
                    )
                )
            )
        )

        for c in range(3):
            if tau == 1:
                e_1_sca = t * (
                    e_theta[a_idx, c] * pilm[l, abs(m), a_idx] * 1j * m
                    - e_phi[a_idx, c] * taulm[l, abs(m), a_idx]
                )
            else:
                e_1_sca = t * (
                    e_phi[a_idx, c] * pilm[l, abs(m), a_idx] * 1j * m
                    + e_theta[a_idx, c] * taulm[l, abs(m), a_idx]
                )

            cuda.atomic.add(e_1_sca_real, (a_idx, c, w_idx), e_1_sca.real)
            cuda.atomic.add(e_1_sca_imag, (a_idx, c, w_idx), e_1_sca.imag)

    @cuda.jit(fastmath=True)
    def compute_electric_field_angle_components(
        lmax: int,
        particles_position: np.ndarray,
        idx: np.ndarray,
        sfc: np.ndarray,
        k_medium: np.ndarray,
        azimuthal_angles: np.ndarray,
        e_r: np.ndarray,
        pilm: np.ndarray,
        taulm: np.ndarray,
        e_field_theta_real: np.ndarray,
        e_field_theta_imag: np.ndarray,
        e_field_phi_real: np.ndarray,
        e_field_phi_imag: np.ndarray,
    ):
        """
        Compute the electric field angle components on the GPU.

        Args:
            lmax (int): The maximum angular momentum quantum number.
            particles_position (np.ndarray): Array of particle positions.
            idx (np.ndarray): Array of indices.
            sfc (np.ndarray): Array of scattering form factors.
            k_medium (np.ndarray): Array of medium wavevectors.
            azimuthal_angles (np.ndarray): Array of azimuthal angles.
            e_r (np.ndarray): Array of radial unit vectors.
            pilm (np.ndarray): Array of associated Legendre polynomials.
            taulm (np.ndarray): Array of tau coefficients.
            e_field_theta_real (np.ndarray): Array of real parts of electric field theta component.
            e_field_theta_imag (np.ndarray): Array of imaginary parts of electric field theta component.
            e_field_phi_real (np.ndarray): Array of real parts of electric field phi component.
            e_field_phi_imag (np.ndarray): Array of imaginary parts of electric field phi component.
        """
        j_idx, a_idx, w_idx = cuda.grid(3)
        jmax = particles_position.shape[0] * 2 * lmax * (lmax + 2)
        if (
            (j_idx >= jmax)
            or (a_idx >= azimuthal_angles.size)
            or (w_idx >= k_medium.size)
        ):
            return

        s, n, tau, l, m = idx[j_idx, :]

        t = (
            1j ** (tau - l - 2)
            * sfc[s, n, w_idx]
            / sqrt(2 * l * (l + 1))
            * exp(
                1j
                * (
                    m * azimuthal_angles[a_idx]
                    - k_medium[w_idx]
                    * (
                        particles_position[s, 0] * e_r[a_idx, 0]
                        + particles_position[s, 1] * e_r[a_idx, 1]
                        + particles_position[s, 2] * e_r[a_idx, 2]
                    )
                )
            )
        )

        if tau == 1:
            e_field_theta = t * pilm[l, abs(m), a_idx] * 1j * m
            e_field_phi = -t * taulm[l, abs(m), a_idx]
        else:
            e_field_theta = t * taulm[l, abs(m), a_idx]
            e_field_phi = t * pilm[l, abs(m), a_idx] * 1j * m

        cuda.atomic.add(e_field_theta_real, (a_idx, w_idx), e_field_theta.real)
        cuda.atomic.add(e_field_theta_imag, (a_idx, w_idx), e_field_theta.imag)
        cuda.atomic.add(e_field_phi_real, (a_idx, w_idx), e_field_phi.real)
        cuda.atomic.add(e_field_phi_imag, (a_idx, w_idx), e_field_phi.imag)

    @cuda.jit(fastmath=True)
    def compute_polarization_components(
        number_of_wavelengths: int,
        number_of_angles: int,
        e_field_theta_real: np.ndarray,
        e_field_theta_imag: np.ndarray,
        e_field_phi_real: np.ndarray,
        e_field_phi_imag: np.ndarray,
        intensity: np.ndarray,
        degree_of_polarization: np.ndarray,
        degree_of_linear_polarization: np.ndarray,
        degree_of_linear_polarization_q: np.ndarray,
        degree_of_linear_polarization_u: np.ndarray,
        degree_of_circular_polarization: np.ndarray,
    ):
        """
        Compute the polarization components using GPU acceleration.

        Args:
            number_of_wavelengths (int): Number of wavelengths.
            number_of_angles (int): Number of angles.
            e_field_theta_real (np.ndarray): Real part of the electric field in the theta direction.
            e_field_theta_imag (np.ndarray): Imaginary part of the electric field in the theta direction.
            e_field_phi_real (np.ndarray): Real part of the electric field in the phi direction.
            e_field_phi_imag (np.ndarray): Imaginary part of the electric field in the phi direction.
            intensity (np.ndarray): Array to store the intensity component.
            degree_of_polarization (np.ndarray): Array to store the degree of polarization component.
            degree_of_linear_polarization (np.ndarray): Array to store the degree of linear polarization component.
            degree_of_linear_polarization_q (np.ndarray): Array to store the degree of linear polarization (Q) component.
            degree_of_linear_polarization_u (np.ndarray): Array to store the degree of linear polarization (U) component.
            degree_of_circular_polarization (np.ndarray): Array to store the degree of circular polarization component.
        """
        a_idx, w_idx = cuda.grid(2)
        if (w_idx >= number_of_wavelengths) or (a_idx >= number_of_angles):
            return

        # Jones vector components (1,2,4)
        e_field_theta_abs = (
            e_field_theta_real[a_idx, w_idx] ** 2
            + e_field_theta_imag[a_idx, w_idx] ** 2
        )
        e_field_phi_abs = (
            e_field_phi_real[a_idx, w_idx] ** 2 + e_field_phi_imag[a_idx, w_idx] ** 2
        )
        e_field_angle_interaction_real = (
            e_field_theta_real[a_idx, w_idx] * e_field_phi_real[a_idx, w_idx]
            + e_field_theta_imag[a_idx, w_idx] * e_field_phi_imag[a_idx, w_idx]
        )
        e_field_angle_interaction_imag = (
            e_field_theta_imag[a_idx, w_idx] * e_field_phi_real[a_idx, w_idx]
            - e_field_theta_real[a_idx, w_idx] * e_field_phi_imag[a_idx, w_idx]
        )

        # Stokes components S = (I, Q, U, V)
        I = e_field_theta_abs + e_field_phi_abs
        Q = e_field_theta_abs - e_field_phi_abs
        U = -2 * e_field_angle_interaction_real
        V = 2 * e_field_angle_interaction_imag

        intensity[a_idx, w_idx] = I
        degree_of_polarization[a_idx, w_idx] = sqrt(Q**2 + U**2 + V**2).real / I
        degree_of_linear_polarization[a_idx, w_idx] = sqrt(Q**2 + U**2).real / I
        degree_of_linear_polarization_q[a_idx, w_idx] = -Q.real / I
        degree_of_linear_polarization_u[a_idx, w_idx] = U.real / I
        degree_of_circular_polarization[a_idx, w_idx] = V / I

    @cuda.jit(fastmath=True)
    def compute_field(
        lmax: int,
        idx: np.ndarray,
        size_parameter: np.ndarray,
        sph_h: np.ndarray,
        derivative: np.ndarray,
        e_j_dm_phi: np.ndarray,
        p_lm: np.ndarray,
        pi_lm: np.ndarray,
        tau_lm: np.ndarray,
        e_r: np.ndarray,
        e_theta: np.ndarray,
        e_phi: np.ndarray,
        scattered_field_coefficients: np.ndarray,
        field_real: np.ndarray,
        field_imag: np.ndarray,
    ):  # , initial_field_coefficients: np.ndarray, scatter_to_internal: np.ndarray):
        """
        Compute the field on the GPU using CUDA.

        Args:
            lmax (int): Maximum degree of the spherical harmonics.
            idx (np.ndarray): Array of indices.
            size_parameter (np.ndarray): Array of size parameters.
            sph_h (np.ndarray): Array of spherical harmonics.
            derivative (np.ndarray): Array of derivatives.
            e_j_dm_phi (np.ndarray): Array of phi-dependent terms.
            p_lm (np.ndarray): Array of Legendre polynomials.
            pi_lm (np.ndarray): Array of pi-dependent terms.
            tau_lm (np.ndarray): Array of tau-dependent terms.
            e_r (np.ndarray): Array of r-dependent terms.
            e_theta (np.ndarray): Array of theta-dependent terms.
            e_phi (np.ndarray): Array of phi-dependent terms.
            scattered_field_coefficients (np.ndarray): Array of scattered field coefficients.
            field_real (np.ndarray): Array to store the real part of the field.
            field_imag (np.ndarray): Array to store the imaginary part of the field.
        """
        jmax = sph_h.shape[1] * 2 * lmax * (lmax + 2)
        channels = sph_h.shape[-1]

        sampling_idx, j_idx, w = cuda.grid(3)

        if (sampling_idx >= sph_h.shape[2]) or (j_idx >= jmax) or (w >= channels):
            return

        particle_idx, n, tau, l, m = idx[j_idx, :]

        invariant = (
            1
            / sqrt(2 * (l + 1) * l)
            * e_j_dm_phi[m + 2 * lmax, particle_idx, sampling_idx]
        )

        for c in range(3):
            term = scattered_field_coefficients[particle_idx, n, w] * invariant

            # Calculate M
            if tau == 1:
                c_term_1 = (
                    pi_lm[l, abs(m), particle_idx, sampling_idx]
                    * e_theta[particle_idx, sampling_idx, c]
                    * 1j
                    * m
                )
                c_term_2 = (
                    tau_lm[l, abs(m), particle_idx, sampling_idx]
                    * e_phi[particle_idx, sampling_idx, c]
                )
                c_term = sph_h[l, particle_idx, sampling_idx, w] * (c_term_1 - c_term_2)

                term *= c_term

            # Calculate N
            else:
                p_term = (
                    l
                    * (l + 1)
                    / size_parameter[particle_idx, sampling_idx, w]
                    * sph_h[l, particle_idx, sampling_idx, w]
                )
                p_term *= (
                    p_lm[l, abs(m), particle_idx, sampling_idx]
                    * e_r[particle_idx, sampling_idx, c]
                )

                b_term_1 = (
                    derivative[l, particle_idx, sampling_idx, w]
                    / size_parameter[particle_idx, sampling_idx, w]
                )
                b_term_2 = (
                    tau_lm[l, abs(m), particle_idx, sampling_idx]
                    * e_theta[particle_idx, sampling_idx, c]
                )
                b_term_3 = (
                    pi_lm[l, abs(m), particle_idx, sampling_idx]
                    * e_phi[particle_idx, sampling_idx, c]
                    * 1j
                    * m
                )
                b_term = b_term_1 * (b_term_2 + b_term_3)

                term *= p_term + b_term

            cuda.atomic.add(field_real, (w, sampling_idx, c), term.real)
            cuda.atomic.add(field_imag, (w, sampling_idx, c), term.imag)



import numpy as np
from numba import cuda
from cmath import exp, sqrt


# TODO: Implement data batching for GPUs with smaller memory
@cuda.jit(fastmath=True)
def particle_interaction_gpu(
    lmax: int,
    particle_number: int,
    idx: np.ndarray,
    x: np.ndarray,
    wx_real: np.ndarray,
    wx_imag: np.ndarray,
    translation_table: np.ndarray,
    plm: np.ndarray,
    sph_h: np.ndarray,
    e_j_dm_phi,
):
    """
    Perform particle interaction calculations on the GPU.

    Args:
        lmax (int): Maximum angular momentum quantum number.
        particle_number (int): Number of particles.
        idx (np.ndarray): Array containing particle indices.
        x (np.ndarray): Array of particle positions.
        wx_real (np.ndarray): Array to store the real part of the result.
        wx_imag (np.ndarray): Array to store the imaginary part of the result.
        translation_table (np.ndarray): Array containing translation table.
        plm (np.ndarray): Array containing associated Legendre polynomials.
        sph_h (np.ndarray): Array containing spherical harmonics.
        e_j_dm_phi (np.ndarray): Additional parameter for the calculation.

    Todo:
        Implement data batching for GPUs with smaller memory
    """
    j1, j2, w = cuda.grid(3)

    jmax = particle_number * 2 * lmax * (lmax + 2)
    channels = sph_h.shape[-1]

    if (j1 >= jmax) or (j2 >= jmax) or (w >= channels):
        return

    s1, n1, tau1, l1, m1 = idx[j1, :]
    s2, n2, tau2, l2, m2 = idx[j2, :]

    if s1 == s2:
        return

    delta_tau = abs(tau1 - tau2)
    delta_l = abs(l1 - l2)
    delta_m = abs(m1 - m2)

    p_dependent = complex(0)
    for p in range(max(delta_m, delta_l + delta_tau), l1 + l2 + 1):
        p_dependent += (
            translation_table[n2, n1, p]
            * plm[p * (p + 1) // 2 + delta_m, s1, s2]
            * sph_h[p, s1, s2, w]
        )
    p_dependent *= e_j_dm_phi[m2 - m1 + 2 * lmax, s1, s2] * x[j2]

    # atomic.add performs the += operation in sync
    cuda.atomic.add(wx_real, (j1, w), p_dependent.real)
    cuda.atomic.add(wx_imag, (j1, w), p_dependent.imag)


@cuda.jit(fastmath=True)
def compute_scattering_cross_section_gpu(
    lmax: int,
    particle_number: int,
    idx: np.ndarray,
    sfc: np.ndarray,
    translation_table: np.ndarray,
    plm: np.ndarray,
    sph_h: np.ndarray,
    e_j_dm_phi: np.ndarray,
    c_sca_real: np.ndarray,
    c_sca_imag: np.ndarray,
):
    """
    Compute the scattering cross section on the GPU using CUDA.

    Args:
        lmax (int): The maximum degree of the spherical harmonics expansion.
        particle_number (int): The number of particles.
        idx (np.ndarray): The index array.
        sfc (np.ndarray): The scattering form factor array.
        translation_table (np.ndarray): The translation table array.
        plm (np.ndarray): The associated Legendre polynomials array.
        sph_h (np.ndarray): The spherical harmonics array.
        e_j_dm_phi (np.ndarray): The phase factor array.
        c_sca_real (np.ndarray): The real part of the scattering cross section array.
        c_sca_imag (np.ndarray): The imaginary part of the scattering cross section array.
    """
    jmax = particle_number * 2 * lmax * (lmax + 2)
    channels = sph_h.shape[-1]

    j1, j2, w = cuda.grid(3)

    if (j1 >= jmax) or (j2 >= jmax) or (w >= channels):
        return

    s1, n1, _, _, m1 = idx[j1, :]
    s2, n2, _, _, m2 = idx[j2, :]

    delta_m = abs(m1 - m2)

    p_dependent = complex(0)
    for p in range(delta_m, 2 * lmax + 1):
        p_dependent += (
            translation_table[n2, n1, p]
            * plm[p * (p + 1) // 2 + delta_m, s1, s2]
            * sph_h[p, s1, s2, w]
        )
    p_dependent *= (
        sfc[s1, n1, w].conjugate()
        * e_j_dm_phi[m2 - m1 + 2 * lmax, s1, s2]
        * sfc[s2, n2, w]
    )

    # atomic.add performs the += operation in sync
    cuda.atomic.add(c_sca_real, w, p_dependent.real)
    cuda.atomic.add(c_sca_imag, w, p_dependent.imag)


@cuda.jit(fastmath=True)
def compute_radial_independent_scattered_field_gpu(
    lmax: int,
    particles_position: np.ndarray,
    idx: np.ndarray,
    sfc: np.ndarray,
    k_medium: np.ndarray,
    azimuthal_angles: np.ndarray,
    e_r: np.ndarray,
    e_phi: np.ndarray,
    e_theta: np.ndarray,
    pilm: np.ndarray,
    taulm: np.ndarray,
    e_1_sca_real: np.ndarray,
    e_1_sca_imag: np.ndarray,
):
    """
    Compute the radial independent scattered field using GPU acceleration.

    Args:
        lmax (int): The maximum degree of the spherical harmonics expansion.
        particles_position (np.ndarray): Array of particle positions.
        idx (np.ndarray): Array of indices for particle properties.
        sfc (np.ndarray): Array of scattering form factors.
        k_medium (np.ndarray): Array of wave numbers in the medium.
        azimuthal_angles (np.ndarray): Array of azimuthal angles.
        e_r (np.ndarray): Array of radial electric field components.
        e_phi (np.ndarray): Array of azimuthal electric field components.
        e_theta (np.ndarray): Array of polar electric field components.
        pilm (np.ndarray): Array of associated Legendre polynomials.
        taulm (np.ndarray): Array of tau coefficients.
        e_1_sca_real (np.ndarray): Array of real parts of the scattered electric field.
        e_1_sca_imag (np.ndarray): Array of imaginary parts of the scattered electric field.
    """
    j_idx, a_idx, w_idx = cuda.grid(3)

    jmax = particles_position.shape[0] * 2 * lmax * (lmax + 2)

    if (j_idx >= jmax) or (a_idx >= azimuthal_angles.size) or (w_idx >= k_medium.size):
        return

    s, n, tau, l, m = idx[j_idx, :]

    # Temporary variable
    # If tau = 1 -> 1j**(tau-1) = 1, if tau = 2 -> 1j**(tau-1) = 1j
    # 1j**(-l-1) = (-1j)**(l+1) => both lead to the coefficient 1j**(tau-l-2)
    # k * <particle_position, e_r> is the phase shift due to the distance and relative position
    t = (
        1j ** (tau - l - 2)
        * sfc[s, n, w_idx]
        / sqrt(2 * l * (l + 1))
        * exp(
            1j
            * (
                m * azimuthal_angles[a_idx]
                - k_medium[w_idx]
                * (
                    particles_position[s, 0] * e_r[a_idx, 0]
                    + particles_position[s, 1] * e_r[a_idx, 1]
                    + particles_position[s, 2] * e_r[a_idx, 2]
                )
            )
        )
    )

    for c in range(3):
        if tau == 1:
            e_1_sca = t * (
                e_theta[a_idx, c] * pilm[l, abs(m), a_idx] * 1j * m
                - e_phi[a_idx, c] * taulm[l, abs(m), a_idx]
            )
        else:
            e_1_sca = t * (
                e_phi[a_idx, c] * pilm[l, abs(m), a_idx] * 1j * m
                + e_theta[a_idx, c] * taulm[l, abs(m), a_idx]
            )

        cuda.atomic.add(e_1_sca_real, (a_idx, c, w_idx), e_1_sca.real)
        cuda.atomic.add(e_1_sca_imag, (a_idx, c, w_idx), e_1_sca.imag)


@cuda.jit(fastmath=True)
def compute_electric_field_angle_components_gpu(
    lmax: int,
    particles_position: np.ndarray,
    idx: np.ndarray,
    sfc: np.ndarray,
    k_medium: np.ndarray,
    azimuthal_angles: np.ndarray,
    e_r: np.ndarray,
    pilm: np.ndarray,
    taulm: np.ndarray,
    e_field_theta_real: np.ndarray,
    e_field_theta_imag: np.ndarray,
    e_field_phi_real: np.ndarray,
    e_field_phi_imag: np.ndarray,
):
    """
    Compute the electric field angle components on the GPU.

    Args:
        lmax (int): The maximum angular momentum quantum number.
        particles_position (np.ndarray): Array of particle positions.
        idx (np.ndarray): Array of indices.
        sfc (np.ndarray): Array of scattering form factors.
        k_medium (np.ndarray): Array of medium wavevectors.
        azimuthal_angles (np.ndarray): Array of azimuthal angles.
        e_r (np.ndarray): Array of radial unit vectors.
        pilm (np.ndarray): Array of associated Legendre polynomials.
        taulm (np.ndarray): Array of tau coefficients.
        e_field_theta_real (np.ndarray): Array of real parts of electric field theta component.
        e_field_theta_imag (np.ndarray): Array of imaginary parts of electric field theta component.
        e_field_phi_real (np.ndarray): Array of real parts of electric field phi component.
        e_field_phi_imag (np.ndarray): Array of imaginary parts of electric field phi component.
    """
    j_idx, a_idx, w_idx = cuda.grid(3)
    jmax = particles_position.shape[0] * 2 * lmax * (lmax + 2)
    if (j_idx >= jmax) or (a_idx >= azimuthal_angles.size) or (w_idx >= k_medium.size):
        return

    s, n, tau, l, m = idx[j_idx, :]

    t = (
        1j ** (tau - l - 2)
        * sfc[s, n, w_idx]
        / sqrt(2 * l * (l + 1))
        * exp(
            1j
            * (
                m * azimuthal_angles[a_idx]
                - k_medium[w_idx]
                * (
                    particles_position[s, 0] * e_r[a_idx, 0]
                    + particles_position[s, 1] * e_r[a_idx, 1]
                    + particles_position[s, 2] * e_r[a_idx, 2]
                )
            )
        )
    )

    if tau == 1:
        e_field_theta = t * pilm[l, abs(m), a_idx] * 1j * m
        e_field_phi = -t * taulm[l, abs(m), a_idx]
    else:
        e_field_theta = t * taulm[l, abs(m), a_idx]
        e_field_phi = t * pilm[l, abs(m), a_idx] * 1j * m

    cuda.atomic.add(e_field_theta_real, (a_idx, w_idx), e_field_theta.real)
    cuda.atomic.add(e_field_theta_imag, (a_idx, w_idx), e_field_theta.imag)
    cuda.atomic.add(e_field_phi_real, (a_idx, w_idx), e_field_phi.real)
    cuda.atomic.add(e_field_phi_imag, (a_idx, w_idx), e_field_phi.imag)


@cuda.jit(fastmath=True)
def compute_polarization_components_gpu(
    number_of_wavelengths: int,
    number_of_angles: int,
    e_field_theta_real: np.ndarray,
    e_field_theta_imag: np.ndarray,
    e_field_phi_real: np.ndarray,
    e_field_phi_imag: np.ndarray,
    intensity: np.ndarray,
    degree_of_polarization: np.ndarray,
    degree_of_linear_polarization: np.ndarray,
    degree_of_linear_polarization_q: np.ndarray,
    degree_of_linear_polarization_u: np.ndarray,
    degree_of_circular_polarization: np.ndarray,
):
    """
    Compute the polarization components using GPU acceleration.

    Args:
        number_of_wavelengths (int): Number of wavelengths.
        number_of_angles (int): Number of angles.
        e_field_theta_real (np.ndarray): Real part of the electric field in the theta direction.
        e_field_theta_imag (np.ndarray): Imaginary part of the electric field in the theta direction.
        e_field_phi_real (np.ndarray): Real part of the electric field in the phi direction.
        e_field_phi_imag (np.ndarray): Imaginary part of the electric field in the phi direction.
        intensity (np.ndarray): Array to store the intensity component.
        degree_of_polarization (np.ndarray): Array to store the degree of polarization component.
        degree_of_linear_polarization (np.ndarray): Array to store the degree of linear polarization component.
        degree_of_linear_polarization_q (np.ndarray): Array to store the degree of linear polarization (Q) component.
        degree_of_linear_polarization_u (np.ndarray): Array to store the degree of linear polarization (U) component.
        degree_of_circular_polarization (np.ndarray): Array to store the degree of circular polarization component.
    """
    a_idx, w_idx = cuda.grid(2)
    if (w_idx >= number_of_wavelengths) or (a_idx >= number_of_angles):
        return

    # Jones vector components (1,2,4)
    e_field_theta_abs = (
        e_field_theta_real[a_idx, w_idx] ** 2 + e_field_theta_imag[a_idx, w_idx] ** 2
    )
    e_field_phi_abs = (
        e_field_phi_real[a_idx, w_idx] ** 2 + e_field_phi_imag[a_idx, w_idx] ** 2
    )
    e_field_angle_interaction_real = (
        e_field_theta_real[a_idx, w_idx] * e_field_phi_real[a_idx, w_idx]
        + e_field_theta_imag[a_idx, w_idx] * e_field_phi_imag[a_idx, w_idx]
    )
    e_field_angle_interaction_imag = (
        e_field_theta_imag[a_idx, w_idx] * e_field_phi_real[a_idx, w_idx]
        - e_field_theta_real[a_idx, w_idx] * e_field_phi_imag[a_idx, w_idx]
    )

    # Stokes components S = (I, Q, U, V)
    I = e_field_theta_abs + e_field_phi_abs
    Q = e_field_theta_abs - e_field_phi_abs
    U = -2 * e_field_angle_interaction_real
    V = 2 * e_field_angle_interaction_imag

    intensity[a_idx, w_idx] = I
    degree_of_polarization[a_idx, w_idx] = sqrt(Q**2 + U**2 + V**2).real / I
    degree_of_linear_polarization[a_idx, w_idx] = sqrt(Q**2 + U**2).real / I
    degree_of_linear_polarization_q[a_idx, w_idx] = -Q.real / I
    degree_of_linear_polarization_u[a_idx, w_idx] = U.real / I
    degree_of_circular_polarization[a_idx, w_idx] = V / I


@cuda.jit(fastmath=True)
def compute_field_gpu(
    lmax: int,
    idx: np.ndarray,
    size_parameter: np.ndarray,
    sph_h: np.ndarray,
    derivative: np.ndarray,
    e_j_dm_phi: np.ndarray,
    p_lm: np.ndarray,
    pi_lm: np.ndarray,
    tau_lm: np.ndarray,
    e_r: np.ndarray,
    e_theta: np.ndarray,
    e_phi: np.ndarray,
    scattered_field_coefficients: np.ndarray,
    field_real: np.ndarray,
    field_imag: np.ndarray,
):  # , initial_field_coefficients: np.ndarray, scatter_to_internal: np.ndarray):
    """
    Compute the field on the GPU using CUDA.

    Args:
        lmax (int): Maximum degree of the spherical harmonics.
        idx (np.ndarray): Array of indices.
        size_parameter (np.ndarray): Array of size parameters.
        sph_h (np.ndarray): Array of spherical harmonics.
        derivative (np.ndarray): Array of derivatives.
        e_j_dm_phi (np.ndarray): Array of phi-dependent terms.
        p_lm (np.ndarray): Array of Legendre polynomials.
        pi_lm (np.ndarray): Array of pi-dependent terms.
        tau_lm (np.ndarray): Array of tau-dependent terms.
        e_r (np.ndarray): Array of r-dependent terms.
        e_theta (np.ndarray): Array of theta-dependent terms.
        e_phi (np.ndarray): Array of phi-dependent terms.
        scattered_field_coefficients (np.ndarray): Array of scattered field coefficients.
        field_real (np.ndarray): Array to store the real part of the field.
        field_imag (np.ndarray): Array to store the imaginary part of the field.
    """
    jmax = sph_h.shape[1] * 2 * lmax * (lmax + 2)
    channels = sph_h.shape[-1]

    sampling_idx, j_idx, w = cuda.grid(3)

    if (sampling_idx >= sph_h.shape[2]) or (j_idx >= jmax) or (w >= channels):
        return

    particle_idx, n, tau, l, m = idx[j_idx, :]

    invariant = (
        1 / sqrt(2 * (l + 1) * l) * e_j_dm_phi[m + 2 * lmax, particle_idx, sampling_idx]
    )

    for c in range(3):
        term = scattered_field_coefficients[particle_idx, n, w] * invariant

        # Calculate M
        if tau == 1:
            c_term_1 = (
                pi_lm[l, abs(m), particle_idx, sampling_idx]
                * e_theta[particle_idx, sampling_idx, c]
                * 1j
                * m
            )
            c_term_2 = (
                tau_lm[l, abs(m), particle_idx, sampling_idx]
                * e_phi[particle_idx, sampling_idx, c]
            )
            c_term = sph_h[l, particle_idx, sampling_idx, w] * (c_term_1 - c_term_2)

            term *= c_term

        # Calculate N
        else:
            p_term = (
                l
                * (l + 1)
                / size_parameter[particle_idx, sampling_idx, w]
                * sph_h[l, particle_idx, sampling_idx, w]
            )
            p_term *= (
                p_lm[l, abs(m), particle_idx, sampling_idx]
                * e_r[particle_idx, sampling_idx, c]
            )

            b_term_1 = (
                derivative[l, particle_idx, sampling_idx, w]
                / size_parameter[particle_idx, sampling_idx, w]
            )
            b_term_2 = (
                tau_lm[l, abs(m), particle_idx, sampling_idx]
                * e_theta[particle_idx, sampling_idx, c]
            )
            b_term_3 = (
                pi_lm[l, abs(m), particle_idx, sampling_idx]
                * e_phi[particle_idx, sampling_idx, c]
                * 1j
                * m
            )
            b_term = b_term_1 * (b_term_2 + b_term_3)

            term *= p_term + b_term

        cuda.atomic.add(field_real, (w, sampling_idx, c), term.real)
        cuda.atomic.add(field_imag, (w, sampling_idx, c), term.imag)



import logging
import yasfpy.log as log

import numpy as np
from scipy.spatial import ConvexHull
from scipy.spatial.distance import pdist

from yasfpy.functions.material_handler import material_handler


class Particles:
    """The `Particles` class represents a collection of particles with various properties such as position,
    radius, and refractive index, and provides methods for computing unique properties and
    characteristics of the particles.
    """

    def __init__(
        self,
        position: np.ndarray,
        r: np.ndarray,
        refractive_index: np.ndarray,
        refractive_index_table: list = None,
        shape_type: str = "sphere",
    ):
        """Initializes an object with position, radius, refractive index, refractive index table, and shape type attributes.

        Args:
            position (np.array): A numpy array representing the position of the shape.
            r (np.array): A numpy array containing the radius values for each shape in the system.
            refractive_index (np.array): A numpy array representing the refractive index of the shape.
                It can be either a complex number or a two-column matrix.
            refractive_index_table (list): A list containing the refractive index values for different materials.
                Each element in the list represents a material, and the refractive index values for that material are stored as a complex number.
            shape_type (str, optional): A string specifying the type of shape for the object.
                Defaults to "sphere" or any other supported shape type.

        """
        self.position = position
        self.r = r
        self.refractive_index = refractive_index
        self.type = shape_type

        # self.log = log.scattering_logger(__name__)
        self.log = logging.getLogger(self.__class__.__module__)

        # TODO: Keep it for now, remove later...
        self.refractive_index_table = refractive_index_table

        if refractive_index_table is None:
            if self.refractive_index.shape[1] == 2:
                self.refractive_index = (
                    self.refractive_index[:, 0] + 1j * self.refractive_index[:, 1]
                )
            elif self.refractive_index.shape[1] > 2:
                self.log.error(
                    "Refractive index should be either complex or a two column matrix!"
                )
        else:
            self.refractive_index = refractive_index.astype(int)
            self.refractive_index_table = refractive_index_table

        self.number = r.shape[0]
        self.__setup_impl()

    @staticmethod
    def generate_refractive_index_table(urls: list) -> list:
        """The function `generate_refractive_index_table` takes a list of URLs, retrieves data from each
        URL using the `material_handler` function, and returns a list of the retrieved data.

        Args:
            urls (list): A list of URLs representing different materials.

        Returns:
            data (list): A list of data. Each element in the list corresponds to a URL in the input list,
                and the data is obtained by calling the `material_handler` function on each URL.

        """
        data = [None] * len(urls)
        for k, url in enumerate(urls):
            data[k] = material_handler(url)

        return data

    def compute_unique_refractive_indices(self):
        """Computes the unique refractive indices and their indices."""
        self.unique_refractive_indices, self.refractive_index_array_idx = np.unique(
            self.refractive_index, return_inverse=True, axis=0
        )
        self.num_unique_refractive_indices = self.unique_refractive_indices.shape[0]

    def compute_unique_radii(self):
        """The function computes the unique radii from an array and stores them in a variable."""
        self.unqiue_radii, self.radius_array_idx = np.unique(
            self.r, return_inverse=True, axis=0
        )
        self.num_unique_radii = self.unqiue_radii.shape[0]

    def compute_unique_radii_index_pairs(self):
        """The function computes unique pairs of radii and refractive indices and stores them in different
        arrays.

        """
        self.unique_radius_index_pairs, self.single_unique_array_idx = np.unique(
            np.column_stack((self.r, self.refractive_index)),
            return_inverse=True,
            axis=0,
        )
        self.unique_single_radius_index_pairs = np.unique(
            np.column_stack((self.radius_array_idx, self.refractive_index_array_idx)),
            axis=0,
        )

    def compute_single_unique_idx(self):
        """The function computes a single unique index based on the sum of pairs of values and their
        corresponding indices.

        """
        self.single_unique_idx = (
            np.sum(self.unique_single_radius_index_pairs, axis=1)
            * (np.sum(self.unique_single_radius_index_pairs, axis=1) + 1)
        ) // 2 + self.unique_single_radius_index_pairs[:, 1]

        # pairedArray = (
        #   self.radius_array_idx + self.refractive_index_array_idx *
        #   (self.radius_array_idx + self.refractive_index_array_idx + 1)
        # ) // 2 + self.refractive_index_array_idx

        # self.single_unique_idx, self.single_unique_array_idx = np.unique(
        #   pairedArray,
        #   return_inverse=True,
        #   axis=0)

        self.num_unique_pairs = self.unique_radius_index_pairs.shape[0]

    def compute_maximal_particle_distance(self):
        """The function computes the maximum distance between particles using the ConvexHull algorithm."""
        if len(self.position) < 4:
            if len(self.position) == 1:
                self.max_particle_distance = 0
            else:
                assert len(self.position) == 1
                "SCIPY CONVEX HULL NEEDS 4 POINT MINIMUM; not currently accounted for"

        else:
            hull = ConvexHull(self.position)
            vert = self.position[hull.vertices, :]
            self.max_particle_distance = max(pdist(vert))

    def compute_volume_equivalent_area(self):
        """The function computes the volume equivalent area by calculating the geometric projection."""
        r3 = np.power(self.r, 3)
        self.geometric_projection = np.pi * np.power(np.sum(r3), 2 / 3)

    def __setup_impl(self):
        """The function sets up various computations related to refractive indices, radii, and particle
        distances.

        """
        self.compute_unique_refractive_indices()
        self.compute_unique_radii()
        self.compute_unique_radii_index_pairs()
        self.compute_single_unique_idx()
        self.compute_maximal_particle_distance()
        self.compute_volume_equivalent_area()



import sys, os

sys.path.append(os.getcwd())
import yasfpy.log as log

from scipy.special import spherical_jn, spherical_yn


def t_entry(tau, l, k_medium, k_sphere, radius, field_type="scattered"):
    """
    Computes an entry in the T Matrix for a given l, k, and tau

    Args:
        tau (float): The value of tau.
        l (int): The value of l.
        k_medium (float): The value of k_medium.
        k_sphere (float): The value of k_sphere.
        radius (float): The value of radius.
        field_type (str, optional): The type of field. Defaults to "scattered".

    Returns:
        (float): The computed entry in the T Matrix.

    Note:
        [scipy.special](https://docs.scipy.org/doc/scipy/reference/generated/scipy.special.spherical_jn.html#scipy.special.spherical_jn){:target="_blank"}
        has also derivative function. Why is it not the same?

        Now:      `djx  = x *  spherical_jn(l-1, x)  - l * jx`<br>
        Possible: `djx  = spherical_jn(l, x, derivative=True)`

    Raises:
        ValueError: If an invalid field type is provided.

    """
    m = k_sphere / k_medium
    x = k_medium * radius
    mx = k_sphere * radius

    jx = spherical_jn(l, x)
    jmx = spherical_jn(l, mx)
    hx = spherical_jn(l, x) + 1j * spherical_yn(l, x)

    djx = x * spherical_jn(l - 1, x) - l * jx
    djmx = mx * spherical_jn(l - 1, mx) - l * jmx
    dhx = x * (spherical_jn(l - 1, x) + 1j * spherical_yn(l - 1, x)) - l * hx

    if (field_type, tau) == ("scattered", 1):
        return -(jmx * djx - jx * djmx) / (jmx * dhx - hx * djmx)  # -b
    if (field_type, tau) == ("scattered", 2):
        return -(m**2 * jmx * djx - jx * djmx) / (m**2 * jmx * dhx - hx * djmx)  # -a
    if (field_type, tau) == ("internal", 1):
        return (jx * dhx - hx * djx) / (jmx * dhx - hx * djmx)  # c
    if (field_type, tau) == ("internal", 2):
        return (m * jx * dhx - m * hx * djx) / (m**2 * jmx * dhx - hx * djmx)  # d
    if (field_type, tau) == ("ratio", 1):
        return (jx * dhx - hx * djx) / -(jmx * djx - jx * djmx)  # c / -b
    if (field_type, tau) == ("ratio", 2):
        return (m * jx * dhx - m * hx * djx) / -(m**2 * jmx * djx - jx * djmx)  # d / -a
    logger = log.scattering_logger("t_entry")
    logger.error("Not a valid field type provided. Returning None!")
    raise ValueError("Not a valid field type provided. Returning None!")



import numpy as np

# https://github.com/disordered-photonics/celes/blob/master/src/mathematics/legendre_normalized_trigon.m


def legendre_normalized_trigon(lmax, x: np.ndarray, y: np.ndarray = None):
    """
    Compute the normalized Legendre polynomials of the first kind for trigonometric arguments.

    Args:
        lmax (int): The maximum degree of the Legendre polynomials.
        x (np.ndarray): The input array of x-coordinates.
        y (np.ndarray, optional): The input array of y-coordinates. Defaults to None.

    Returns:
        plm (np.ndarray): The array of computed Legendre polynomials.

    Note:
        Base on the celes implementation of [legendre_normalized_trigon.m](https://github.com/disordered-photonics/celes/blob/master/src/mathematics/legendre_normalized_trigon.m){:target="_blank"}

    Examples:
        >>> lmax = 2
        >>> x = np.array([0, np.pi/4, np.pi/2])
        >>> y = np.array([0, 1, 0])
        >>> result = legendre_normalized_trigon(lmax, x, y)
        >>> print(result)
        array([[[ 0.70710678,  0.        ,  0.        ],
                [ 0.        ,  0.        ,  0.        ],
                [ 0.        ,  0.        ,  0.        ]],
               [[ 0.        ,  0.        ,  0.        ],
                [ 0.        ,  0.70710678,  0.        ],
                [ 0.        ,  0.        ,  0.        ]],
               [[ 0.        ,  0.        ,  0.        ],
                [ 0.        ,  0.        ,  0.        ],
                [ 0.        ,  0.        ,  0.70710678]]])
    """
    #   if not(isinstance(x, np.ndarray) or np.isscalar(x) or isinstance(x, list)):
    #     return legendre_normalized_trigon_legacy(x, y, lmax)

    #   if np.isscalar(x):
    #     x = np.array([x])
    #   elif isinstance(x, list):
    #     x = np.array(x)
    size = x.shape
    if y is None:
        x = np.ravel(x)
        ct = np.cos(x)
        st = np.sin(x)
    else:
        ct = x.ravel()
        st = y.ravel()

    plm = np.zeros((lmax + 1, lmax + 1, ct.size)) * np.nan

    plm[0, 0, :] = np.sqrt(1 / 2) * np.ones_like(ct)
    plm[1, 0, :] = np.sqrt(3 / 2) * ct

    for l in range(1, lmax):
        plm[l + 1, 0, :] = (
            1 / (l + 1) * np.sqrt((2 * l + 1) * (2 * l + 3)) * plm[l, 0, :] * ct
            - l / (l + 1) * np.sqrt((2 * l + 3) / (2 * l - 1)) * plm[l - 1, 0, :]
        )

    for m in range(1, lmax + 1):
        plm[m - 1, m, :] = np.zeros_like(ct)
        plm[m, m, :] = (
            np.sqrt((2 * m + 1) / 2 / np.math.factorial(2 * m))
            * np.prod(np.arange(1, 2 * m, 2))
            * np.power(st, m)
        )
        for l in range(m, lmax):
            plm[l + 1, m, :] = (
                np.sqrt((2 * l + 1) * (2 * l + 3) / (l + 1 - m) / (l + 1 + m))
                * ct
                * plm[l, m, :]
                - np.sqrt(
                    (2 * l + 3)
                    * (l - m)
                    * (l + m)
                    / (2 * l - 1)
                    / (l + 1 - m)
                    / (l + 1 + m)
                )
                * plm[l - 1, m, :]
            )

    plm = np.reshape(plm, np.concatenate(([lmax + 1, lmax + 1], size)))

    return plm


def legendre_normalized_trigon_legacy(x, y=None, lmax=4):
    """
    Calculate the normalized Legendre polynomials for trigonometric functions.

    Args:
        x (float or sympy.core.symbol.Symbol): The input variable x.
        y (float or sympy.core.symbol.Symbol, optional): The input variable y. Defaults to None.
        lmax (int, optional): The maximum degree of the Legendre polynomials. Defaults to 4.

    Returns:
        plm (numpy.ndarray): The matrix of Legendre polynomials.

    """
    import sympy as sym

    plm = sym.zeros(lmax + 1, lmax + 1)
    if y is None:
        ct = sym.cos(x)
        st = sym.sin(x)
    elif isinstance(x, sym.core.symbol.Symbol) and isinstance(
        y, sym.core.symbol.Symbol
    ):
        ct = x
        st = y
    else:
        ct = sym.Symbol("ct")
        st = sym.Symbol("st")

    plm[0, 0] = np.sqrt(1 / 2)
    plm[1, 0] = np.sqrt(3 / 2) * ct

    for l in range(1, lmax):
        plm[l + 1, 0] = (
            1 / (l + 1) * np.sqrt((2 * l + 1) * (2 * l + 3)) * plm[l, 0] * ct
            - l / (l + 1) * np.sqrt((2 * l + 3) / (2 * l - 1)) * plm[l - 1, 0]
        )

    for m in range(1, lmax + 1):
        plm[m - 1, m] = np.zeros_like(ct)
        plm[m, m] = (
            np.sqrt((2 * m + 1) / 2 / np.math.factorial(2 * m))
            * np.prod(np.arange(1, 2 * m, 2))
            * st**m
        )
        for l in range(m, lmax):
            plm[l + 1, m] = (
                np.sqrt((2 * l + 1) * (2 * l + 3) / (l + 1 - m) / (l + 1 + m))
                * ct
                * plm[l, m]
                - np.sqrt(
                    (2 * l + 3)
                    * (l - m)
                    * (l + m)
                    / (2 * l - 1)
                    / (l + 1 - m)
                    / (l + 1 + m)
                )
                * plm[l - 1, m]
            )

    return plm



import numpy as np
import math


def spherical_functions_trigon(lmax, theta, st=None):
    """
    Compute spherical functions using trigonometric functions.

    Args:
        lmax (int): The maximum degree of the spherical harmonics.
        theta (numpy.ndarray or float): The polar angle(s) in radians.
        st (numpy.ndarray or float, optional): The sine of the polar angle(s). If not provided, it will be computed from theta.

    Returns:
        pilm (numpy.ndarray): The associated Legendre functions.
        taulm (numpy.ndarray): The derivative of the associated Legendre functions.

    """
    size = np.array([1])
    if isinstance(theta, np.ndarray):
        size = theta.shape

    if st is None:
        ct = np.cos(theta)
        st = np.sin(theta)
    else:
        ct = np.array(theta)
        st = np.array(st)

    ct = ct.ravel()
    st = st.ravel()

    plm = np.zeros((lmax + 1, lmax + 1, ct.shape[0])) * np.nan
    pilm = np.zeros(plm.shape) * np.nan
    taulm = np.zeros(plm.shape) * np.nan
    pprimel0 = np.zeros((lmax + 1, ct.shape[0])) * np.nan

    plm[0, 0, :] = np.sqrt(1 / 2) * np.ones_like(ct)
    plm[1, 0, :] = np.sqrt(3 / 2) * ct

    pilm[0, 0, :] = np.zeros_like(ct)
    pilm[1, 0, :] = np.zeros_like(ct)

    pprimel0[0, :] = np.zeros_like(ct)
    pprimel0[1, :] = np.sqrt(3) * plm[0, 0, :]

    taulm[0, 0, :] = -st * pprimel0[0, :]
    taulm[1, 0, :] = -st * pprimel0[1, :]

    for l in range(1, lmax):
        plm[l + 1, 0, :] = (
            1 / (l + 1) * np.sqrt((2 * l + 1) * (2 * l + 3)) * ct * plm[l, 0, :]
            - l / (l + 1) * np.sqrt((2 * l + 3) / (2 * l - 1)) * plm[l - 1, 0, :]
        )
        pilm[l + 1, 0, :] = np.zeros_like(ct)
        pprimel0[l + 1, :] = np.sqrt((2 * l + 3) / (2 * l + 1)) * (
            (l + 1) * plm[l, 0, :] + ct * pprimel0[l, :]
        )
        taulm[l + 1, 0, :] = -st * pprimel0[l + 1, :]

    for m in range(1, lmax + 1):
        plm[m - 1, m, :] = np.zeros_like(ct)
        pilm[m - 1, m, :] = np.zeros_like(ct)
        coeff = np.sqrt((2 * m + 1) / 2 / math.factorial(2 * m)) * np.prod(
            np.arange(1, 2 * m, 2)
        )
        plm[m, m, :] = coeff * np.power(st, m)
        pilm[m, m, :] = coeff * np.power(st, m - 1)
        taulm[m, m, :] = m * ct * pilm[m, m, :]
        for l in range(m, lmax):
            coeff1 = np.sqrt((2 * l + 1) * (2 * l + 3) / (l + 1 - m) / (l + 1 + m)) * ct
            coeff2 = np.sqrt(
                (2 * l + 3)
                * (l - m)
                * (l + m)
                / (2 * l - 1)
                / (l + 1 - m)
                / (l + 1 + m)
            )
            plm[l + 1, m, :] = coeff1 * plm[l, m, :] - coeff2 * plm[l - 1, m, :]
            pilm[l + 1, m, :] = coeff1 * pilm[l, m, :] - coeff2 * pilm[l - 1, m, :]
            taulm[l + 1, m, :] = (l + 1) * ct * pilm[l + 1, m, :] - (
                l + 1 + m
            ) * np.sqrt((2 * l + 3) * (l + 1 - m) / (2 * l + 1) / (l + 1 + m)) * pilm[
                l, m, :
            ]

    pilm = np.reshape(pilm, np.concatenate(([lmax + 1, lmax + 1], size)))
    taulm = np.reshape(taulm, np.concatenate(([lmax + 1, lmax + 1], size)))
    return pilm, taulm
